/* -*- Mode: C++ -*-
 * $Id: Sensors.cc,v 1.33 2002/09/26 15:23:41 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Sensors.cc
 * Unitname: BasicSystem
 * $Revision: 1.33 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: s02davro $ $Date: 2002/09/26 15:23:41 $
 *
 *
 * HISTORY
 *
 * $Log: Sensors.cc,v $
 * Revision 1.33  2002/09/26 15:23:41  s02davro
 * Bugfixes, close, team, coach
 *
 * Revision 1.32  2002/09/23 10:28:41  s02davro
 * Find bug in (close), this information needs to be handled - todo
 *
 * Revision 1.31  2002/09/21 15:38:57  s02davro
 * Update to new (1.0.3) RCCParser
 *
 * Revision 1.30  2002/08/31 15:06:16  s02davro
 * Redesign (again) now handles the coach fullstate information messages
 *
 * Revision 1.29  2002/08/13 10:23:17  s02davro
 * Update to handle ServerParam, PlayerParam and PlayerType messages.
 *
 * Revision 1.28  2002/08/02 21:51:49  s02davro
 * Complete rewrite, now uses rccparser-1.0 and clangparser from the soccerserver.
 *
 * Revision 1.26  2002/01/20 17:16:49  frehe
 * Fixed the problem with the coach not hearing the referee messages.
 *
 * Revision 1.25  2002/01/16 09:53:52  frehe
 * Fixed the coach parsing of referee messages.
 *
 * Revision 1.24  2002/01/15 17:19:41  frehe
 * Improved the coach parser.
 *
 * Revision 1.23  2002/01/11 15:27:18  frehe
 * Added support for coach freeform messages.
 *
 * Revision 1.22  2001/11/15 16:48:48  frehe
 * Added support for a coach.
 *
 * Revision 1.21  2001/10/14 21:21:43  frehe
 * Added better support for messages from self and from other players.
 *
 * Revision 1.20  2001/10/14 14:48:36  frehe
 * Removed old debug info.
 *
 * Revision 1.19  2001/09/22 17:18:28  frehe
 * Added a workaround to get teammates from rccparser.
 *
 * Revision 1.18  2001/09/17 21:53:59  frehe
 * Readded support for a debug coach.
 *
 * Revision 1.17  2001/09/17 21:30:09  andbo928
 * Implemented an ugly workaround to get ballspeeds working.
 *
 * Revision 1.16  2001/09/14 13:48:04  frehe
 * Fixed the speed bug.
 *
 * Revision 1.15  2001/09/03 19:39:36  frehe
 * Changed code to use the RS_BEGIN_TRY, RS_END_TRY and RS_CATCH macros.
 *
 * Revision 1.14  2001/09/01 23:52:23  frehe
 * Added some debug info.
 *
 * Revision 1.13  2001/08/30 21:41:47  frehe
 * Removed a call to distDirChng in RCCParser that crashed the agent.
 *
 * Revision 1.12  2001/08/29 16:37:32  frehe
 * Added a test on the validity of vec in getVectorUnknown(RCC_Vector2d).
 *
 * Revision 1.11  2001/08/29 12:35:26  frehe
 * Fixed a bug in sense_body handler.
 *
 * Revision 1.10  2001/08/22 21:06:59  frehe
 * Updated 2001-08-22
 *
 * Revision 1.9  2001/08/19 23:32:24  frehe
 * Updated 2001-08-19
 *
 * Revision 1.8  2001/08/17 07:15:16  frehe
 * Made changes to compile RoboSoc with RCCParser-0.10.
 *
 * Revision 1.7  2001/08/13 16:43:16  frehe
 * Made updates to compile with new autotools and RCCParser.
 *
 * Revision 1.6  2001/08/13 07:33:16  frehe
 * Updated from laptop 2001-08-13.
 *
 * Revision 1.5  2001/08/02 17:04:22  frehe
 * Updated 2001-08-02
 *
 * Revision 1.4  2000/10/24 15:15:17  frehe
 * Fixed the generation of the onNewCycle event.
 *
 * Revision 1.3  2000/09/05 21:15:53  frehe
 * Made minor changes.
 *
 * Revision 1.2  2000/09/04 08:46:45  frehe
 * Fixed a naming error in the sensors.
 *
 * Revision 1.1  2000/09/03 20:41:14  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Sensors.cc
 * Implementation file for the class Sensors.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.33 $
 * $Author: s02davro $
 * $Date: 2002/09/26 15:23:41 $
 */


#include <sstream>
#include <cstdio>
#include <cstring>
#include <cctype>

#include "Sensors.h"
#include "Controller.h"
#include "SensorInterface.h"
#include "string_utils.h"
#include "rs_error.h"
#include "rs_log.h"
#include "rs_try_catch.h"

RS_BEGIN_NAMESPACE


Sensors::Sensors(SensorInterface* sensor_interface, 
		 rcss::Parser& coachMessageParser)
  : rcc::Parser(coachMessageParser),
    facts(WorldFacts::nonConstInstance()),
    theController(NULL),
    theSensorInterface(sensor_interface),
    lastSensorType(SENSOR_UNKNOWN),
    currentTime(0),
    commandQueue(),
    parsedSide(S_UNKNOWN),
    parsedPlayMode(PM_UNKNOWN),
    parsedViewQuality(VQ_UNKNOWN),
    parsedViewWidth(VW_UNKNOWN),
    parsedLocations(),
    parsedFlagOffset(UIntUnknown()),
    parsedUniformNumber(UIntUnknown()),
    parsedGoalie(false),
    parsedTeam(T_UNKNOWN) {
}

void Sensors::init(Controller* controller) {
  theController = controller;
}

void Sensors::analyze(ServerCommand* command) {
  lastSensorType = SENSOR_ACTUATORS;
  commandQueue.push(command->getType());
  theSensorInterface->setLastCommand(command);
  theSensorInterface->setLastSensorType(lastSensorType);  
}

void Sensors::analyze(char* sensor_data) {
  lastSensorType = SENSOR_UNKNOWN;
  std::istrstream strm(sensor_data);
  if (!parse(strm))
    RS_LOG ( LA_SENSORS, "Parse error: " << sensor_data);  
  else if ( lastSensorType == SENSOR_UNKNOWN ) 
    RS_LOG( LA_SENSORS, "UNKNOWN: " << sensor_data );
  else 
    theSensorInterface->setLastSensorType(lastSensorType);
}

void Sensors::updateTime(int time) {
  setCurrentTime(time);
  if ( theController->updateTime(getCurrentTime()) ) {
    theSensorInterface->newCycle();
    RS_LOG( LA_CONTROLLER, "Time: " << toString(getCurrentTime()) 
	    << " --- NEW CYCLE ---");
  }
}

void Sensors::doMsgParsed() {
  RS_LOG( LA_SENSORS, "MsgParsed");
}

void Sensors::doParsedCLang( const std::string& msg ) {
  RS_LOG ( LA_SENSORS, "ParsedCLang: " << msg);
  lastSensorType = SENSOR_AURAL_COACH;
}

void Sensors::doBuildFullState( int time ) {
  RS_LOG ( LA_SENSORS, "BuildFullState: " << time);
  lastSensorType = SENSOR_COACH;
  updateTime(time);
}

void Sensors::doBuildCatchBallLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildCatchBallLeftPlayMode");
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_GOALIE_GOT_BALL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_GOALIE_GOT_BALL;
}

void Sensors::doBuildCatchBallRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildCatchBallRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_GOALIE_GOT_BALL;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_GOALIE_GOT_BALL;
}

void Sensors::doBuildBeforeKickOffPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildBeforeKickOffPlayMode" );
  parsedPlayMode = PM_BEFORE_KICK_OFF;
}

void Sensors::doBuildTimeOverPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildTimeOverPlayMode" );
  parsedPlayMode = PM_TIME_OVER;
}

void Sensors::doBuildPlayOnPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildPlayOnPlayMode" );
  parsedPlayMode = PM_PLAY_ON;
}

void Sensors::doBuildKickOffLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildKickOffLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_KICK_OFF;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_KICK_OFF;
}

void Sensors::doBuildKickOffRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildKickOffRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_KICK_OFF;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_KICK_OFF;
}
 
void Sensors::doBuildKickInLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildKickInLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_KICK_IN;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_KICK_IN;
}
 
void Sensors::doBuildKickInRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildKickInRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_KICK_IN;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_KICK_IN;
}

void Sensors::doBuildFreeKickLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFreeKickLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FREE_KICK;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FREE_KICK;  
}

void Sensors::doBuildFreeKickRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFreeKickRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_FREE_KICK;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_FREE_KICK;  
}

void Sensors::doBuildCornerKickLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildCornerKickLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_CORNER_KICK;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_CORNER_KICK;  
}

void Sensors::doBuildCornerKickRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildCornerKickRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_CORNER_KICK;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_CORNER_KICK;  
}

void Sensors::doBuildGoalKickLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildGoalKickLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_GOAL_KICK;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_GOAL_KICK;  
}

void Sensors::doBuildGoalKickRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildGoalKickRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_GOAL_KICK;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_GOAL_KICK;  
}

void Sensors::doBuildAfterGoalLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildAfterGoalLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT) 
    parsedPlayMode = PM_OUR_GOAL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_GOAL;
}
 
void Sensors::doBuildAfterGoalRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildAfterGoalRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT) 
    parsedPlayMode = PM_OUR_GOAL;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_GOAL;
}

void Sensors::doBuildDropBallPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildDropBallPlayMode" );
  parsedPlayMode = PM_DROP_BALL;
}

void Sensors::doBuildOffSideLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildOffSideLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_OFFSIDE_KICK;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_OFFSIDE_KICK;
}

void Sensors::doBuildOffsideRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildOffsideRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_OFFSIDE_KICK;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_OFFSIDE_KICK;
}
 
void Sensors::doBuildPenaltyKickLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildPenaltyKickLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_PENALTY_KICK;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_PENALTY_KICK;
}

void Sensors::doBuildPenaltyKickRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildPenaltyKickRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_PENALTY_KICK;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_PENALTY_KICK;
}
 
void Sensors::doBuildFirstHalfOverPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFirstHalfOverPlayMode" );
  parsedPlayMode = PM_HALF_TIME;
}

void Sensors::doBuildPausePlayMode() {
  RS_LOG ( LA_SENSORS, "BuildPausePlayMode" );
  parsedPlayMode = PM_PAUSE;
}

void Sensors::doBuildHumanPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildHumanPlayMode" );
  parsedPlayMode = PM_HUMAN;
}
 
void Sensors::doBuildFoulChargeLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulChargeLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFoulChargeRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulChargeRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFoulPushLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulPushLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFoulPushRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulPushRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFoulMultipleAttackerLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulMultipleAttackerLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFoulMultipleAttackerRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulMultipleAttackerRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFoulBallOutLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulBallOutLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFoulBallOutRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulBallOutRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildBackPassLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildBackPassLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildBackPassRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildBackPassRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFreeKickFaultLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFreeKickFaultLeftPlayMode" );
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildFreeKickFaultRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFreeKickFaultRightPlayMode" );
  if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildScore( int our, int opp ) {
  RS_LOG ( LA_SENSORS, "BuildScore: " << our << ", " << opp );
  facts->setScore(facts->getOurSide(), our);
  facts->setScore(facts->getTheirSide(), opp);
  lastSensorType = SENSOR_AURAL_SCORE;
}

void Sensors::doBuildLeftSide() {
  RS_LOG ( LA_SENSORS, "BuildLeftSide" );
  parsedSide = S_LEFT;
}

void Sensors::doBuildRightSide() {
  RS_LOG ( LA_SENSORS, "BuildRightSide" );
  parsedSide = S_RIGHT;
}

void Sensors::doBuildOppSide() {
  RS_LOG ( LA_SENSORS, "BuildOppSide" );
  if (facts->getOurSide() == S_LEFT)
    parsedSide = S_RIGHT;
  else if (facts->getOurSide() == S_RIGHT)
    parsedSide = S_LEFT;
}
 
void Sensors::doBuildOurSide() {
  RS_LOG ( LA_SENSORS, "BuildOurSide" );
  parsedSide = facts->getOurSide();
}
 
void Sensors::doBuildNeutralSide() {
  RS_LOG ( LA_SENSORS, "BuildNeutralSide" );
  parsedSide = S_NEUTRAL;  
}

void Sensors::doBuildSenseBody( int time, 
				double stamina,
				double effort,
				double speed_mag,
				double speed_head,
				double head_angle,
				int kick_count,
				int dash_count,
				int turn_count,
				int say_count,
				int turn_neck_count,
				int catch_count,
				int move_count,
				int chg_view_count ) {
  RS_LOG ( LA_SENSORS, "BuildSenseBody: " << time  << ", " << stamina
	   << ", " << effort << ", " << speed_mag  << ", " << speed_head
	   << ", " << head_angle << ", " << kick_count << ", " << dash_count
	   << ", " << turn_count << ", " << say_count << ", "
	   << turn_neck_count  << ", " << catch_count  << ", " << move_count
	   << ", " << chg_view_count);
  lastSensorType = SENSOR_SENSE;
  
  // @todo Move some things away to better place.

  static UInt catches      = 0;
  static UInt change_views = 0;
  static UInt dashes       = 0;
  static UInt kicks        = 0;
  static UInt moves        = 0;
  static UInt says         = 0;
  static UInt turns        = 0;
  static UInt turn_necks   = 0;

  Int old_catches = catches;
  Int old_change_views = change_views;
  Int old_dashes = dashes;
  Int old_kicks = kicks;
  Int old_moves = moves;
  Int old_says = says;
  Int old_turns = turns;
  Int old_turn_necks = turn_necks;

  catches = catch_count;
  change_views = chg_view_count;
  dashes = dash_count;
  kicks = kick_count;
  moves = move_count;
  says = say_count;
  turns = turn_count;
  turn_necks = turn_neck_count;
    
  old_catches      = catches - old_catches;
  old_change_views = change_views - old_change_views;
  old_dashes       = dashes - old_dashes;
  old_kicks        = kicks - old_kicks;
  old_moves        = moves - old_moves;
  old_says         = says - old_says;
  old_turns        = turns - old_turns;
  old_turn_necks   = turn_necks - old_turn_necks;

  while (!commandQueue.empty()) {
    switch (commandQueue.front()) {
    case CMD_Catch:
      --old_catches;
      break;
    case CMD_ChangeView:
      --old_change_views;
      break;
    case CMD_Dash:
      --old_dashes;
      break;
    case CMD_Kick:
      --old_kicks;
      break;
    case CMD_Move:
      --old_moves;
      break;
    case CMD_Say:
      --old_says;
      break;
    case CMD_Turn:
      --old_turns;
      break;
    case CMD_TurnNeck:
      --old_turn_necks;
      break;
    default:
      //CMD_Unknown, CMD_Bye, CMD_Init, CMD_SenseBody
      break;
    }
    commandQueue.pop();
  }

  if ( old_catches < 0 || old_change_views < 0 ||
       old_dashes < 0 || old_kicks < 0 ||
       old_moves < 0 || old_says < 0 ||
       old_turns < 0 || old_turn_necks < 0 )
    theController->delayedActions(old_catches, old_change_views,
				  old_dashes, old_kicks, old_moves,
				  old_says, old_turns, old_turn_necks);

  if ( old_catches > 0 || old_change_views > 0 ||
       old_dashes > 0 || old_kicks > 0 ||
       old_moves > 0 || old_says > 0 ||
       old_turns > 0 || old_turn_necks > 0 )
    theController->missingActions(old_catches, old_change_views,
				  old_dashes, old_kicks, old_moves,
				  old_says, old_turns, old_turn_necks);


  updateTime(time);
  parsedAgent = ServerAgentObject 
    (parsedViewQuality, 
     parsedViewWidth,
     stamina,
     effort,
     FloatUE(ERROR_QUANTIZE, speed_mag, 0.01),
     AngleDegUE(ERROR_QUANTIZE, speed_head, facts->SP_angle_qstep),
     AngleDegUE(ERROR_QUANTIZE, head_angle, facts->SP_angle_qstep),
     catch_count, 
     chg_view_count, 
     dash_count,
     kick_count, 
     move_count, 
     say_count,
     turn_count, 
     turn_neck_count);  
  theSensorInterface->sense(parsedAgent);
  parsedViewQuality = VQ_UNKNOWN;
  parsedViewWidth = VW_UNKNOWN;  
}

void Sensors::doBuildViewMode() {
  RS_LOG ( LA_SENSORS, "BuildViewMode" );
}

void Sensors::doBuildViewQualityHigh() {
  RS_LOG ( LA_SENSORS, "BuildViewQualityHigh" );
  parsedViewQuality = VQ_HIGH;
}

void Sensors::doBuildViewQualityLow() {
  RS_LOG ( LA_SENSORS, "BuildViewQualityLow" );
  parsedViewQuality = VQ_LOW;
}

void Sensors::doBuildViewWidthNarrow() {
  RS_LOG ( LA_SENSORS, "BuildViewWidthNarrow" );
  parsedViewWidth = VW_NARROW;
}

void Sensors::doBuildViewWidthNormal() {
  RS_LOG ( LA_SENSORS, "BuildViewWidthNormal" );
  parsedViewWidth = VW_NORMAL;
}

void Sensors::doBuildViewWidthWide() {
  RS_LOG ( LA_SENSORS, "BuildViewWidthWide" );
  parsedViewWidth = VW_WIDE;
}

void Sensors::doBuildServerParam ( double gwidth, 
				   double inertia_moment, 
				   double psize, 
				   double pdecay, 
				   double prand,
				   double pweight, 
				   double pspeed_max, 
				   double paccel_max, 
				   double stamina_max, 
				   double stamina_inc,
				   double recover_init, 
				   double recover_dthr, 
				   double recover_min, 
				   double recover_dec, 
				   double effort_init,
				   double effort_dthr, 
				   double effort_min, 
				   double effort_dec, 
				   double effort_ithr, 
				   double effort_inc,
				   double kick_rand, 
				   int team_actuator_noise, 
				   double prand_factor_l, 
				   double prand_factor_r, 
				   double kick_rand_factor_l,
				   double kick_rand_factor_r, 
				   double bsize, 
				   double bdecay, 
				   double brand, 
				   double bweight,
				   double bspeed_max, 
				   double baccel_max, 
				   double dprate, 
				   double kprate, 
				   double kmargin,
				   double ctlradius, 
				   double ctlradius_width, 
				   double maxp, 
				   double minp, 
				   double maxm,
				   double minm, 
				   double maxnm, 
				   double minnm, 
				   double maxn, 
				   double minn,
				   double visangle, 
				   double visdist, 
				   double windir, 
				   double winforce, 
				   double winang,
				   double winrand, 
				   double kickable_area, 
				   double catch_area_l, 
				   double catch_area_w, 
				   double catch_prob,
				   int goalie_max_moves, 
				   double ckmargin, 
				   double offside_area, 
				   int win_no, 
				   int win_random,
				   int say_cnt_max, 
				   int SayCoachMsgSize, 
				   int clang_win_size, 
				   int clang_define_win, 
				   int clang_meta_win,
				   int clang_advice_win, 
				   int clang_info_win, 
				   int clang_mess_delay, 
				   int clang_mess_per_cycle, 
				   int half_time,
				   int sim_st, 
				   int send_st, 
				   int recv_st, 
				   int sb_step, 
				   int lcm_st,
				   int SayMsgSize, 
				   int hear_max, 
				   int hear_inc, 
				   int hear_decay, 
				   int cban_cycle,
				   int slow_down_factor, 
				   int useoffside, 
				   int kickoffoffside, 
				   double offside_kick_margin, 
				   double audio_dist,
				   double dist_qstep, 
				   double land_qstep, 
				   double dir_qstep, 
				   double dist_qstep_l, 
				   double dist_qstep_r,
				   double land_qstep_l, 
				   double land_qstep_r, 
				   double dir_qstep_l, 
				   double dir_qstep_r, 
				   int CoachMode,
				   int CwRMode, 
				   int old_hear, 
				   int sv_st, 
				   int start_goal_l, 
				   int start_goal_r,
				   int fullstate_l, 
				   int fullstate_r, 
				   int drop_time ) {
  RS_LOG ( LA_SENSORS, "BuildServerParam" );
  facts->SP_goal_width = gwidth;
  facts->SP_inertia_moment = inertia_moment;
  facts->SP_player_size = psize;
  facts->SP_player_decay = pdecay;
  facts->SP_player_rand = prand;
  facts->SP_player_weight = pweight;
  facts->SP_player_speed_max = pspeed_max;
  facts->SP_player_accel_max = paccel_max;
  facts->SP_stamina_max = stamina_max;
  facts->SP_stamina_inc_max = stamina_inc;    
  facts->SP_recover_max = recover_init;
  facts->SP_recover_dec_thr = recover_dthr;
  facts->SP_recover_min = recover_min;
  facts->SP_recover_dec = recover_dec;
  facts->SP_effort_max = effort_init;
  facts->SP_effort_dec_thr = effort_dthr;
  facts->SP_effort_min = effort_min;
  facts->SP_effort_dec = effort_dec;
  facts->SP_effort_inc_thr = effort_ithr;
  facts->SP_effort_inc = effort_inc;
  facts->SP_kick_rand = kick_rand;
  facts->SP_team_actuator_noise = team_actuator_noise;
  facts->SP_player_rand_factor_l = prand_factor_l;
  facts->SP_player_rand_factor_r = prand_factor_r;
  facts->SP_kick_rand_factor_l = kick_rand_factor_l;
  facts->SP_kick_rand_factor_r = kick_rand_factor_r;
  facts->SP_ball_size = bsize;
  facts->SP_ball_decay = bdecay;
  facts->SP_ball_rand = brand;
  facts->SP_ball_weight = bweight;
  facts->SP_ball_speed_max = bspeed_max;
  facts->SP_ball_accel_max = baccel_max;
  facts->SP_dash_power_rate = dprate;
  facts->SP_kick_power_rate = kprate;
  facts->SP_kickable_margin = kmargin;
  facts->SP_control_radius = ctlradius;
  facts->SP_control_radius_width = ctlradius_width;
  facts->SP_maxpower = maxp;
  facts->SP_minpower = minp;
  facts->SP_maxmoment = maxm;
  facts->SP_minmoment = minm;
  facts->SP_maxneckmoment = maxnm;
  facts->SP_minneckmoment = minnm;
  facts->SP_maxneckang = maxn;
  facts->SP_minneckang = minn;
  facts->SP_visible_angle = visangle;
  facts->SP_feel_distance = visdist;
  facts->SP_wind_dir = windir;
  facts->SP_wind_force = winforce;
  facts->SP_wind_ang = winang;
  facts->SP_wind_rand = winrand;
  facts->SP_kickable_area = kickable_area;
  facts->SP_catchable_area_l = catch_area_l;
  facts->SP_catchable_area_w = catch_area_w;
  facts->SP_catch_probability = catch_prob;
  facts->SP_goalie_max_moves = goalie_max_moves;
  facts->SP_cornerkick_margin = ckmargin;
  facts->SP_offside_active_area_size = offside_area;
  facts->SP_wind_none = win_no;
  facts->SP_wind_random = win_random;
  facts->SP_say_coach_cnt_max = say_cnt_max;
  facts->SP_say_coach_msg_size = SayCoachMsgSize;
  facts->SP_clang_win_size = clang_win_size;
  facts->SP_clang_define_win = clang_define_win;
  facts->SP_clang_meta_win = clang_meta_win;
  facts->SP_clang_advice_win = clang_advice_win;
  facts->SP_clang_info_win = clang_info_win;
  facts->SP_clang_mess_delay = clang_mess_delay;
  facts->SP_clang_mess_per_cycle = clang_mess_per_cycle;
  facts->SP_half_time = half_time;
  facts->SP_simulator_step = sim_st;
  facts->SP_send_step = send_st;
  facts->SP_recv_step = recv_st;
  facts->SP_sense_body_step = sb_step;
  facts->SP_lcm_step = lcm_st;
  facts->SP_say_msg_size = SayMsgSize;
  facts->SP_hear_max = hear_max;
  facts->SP_hear_inc = hear_inc;
  facts->SP_hear_decay = hear_decay;
  facts->SP_catch_ban_cycle = cban_cycle;
  facts->SP_slow_down_factor = slow_down_factor;
  facts->SP_use_offside = useoffside;
  facts->SP_forbid_kick_off_offside = kickoffoffside;
  facts->SP_offside_kick_margin = offside_kick_margin;
  facts->SP_audio_cut_dist = audio_dist;
  facts->SP_dist_qstep = dist_qstep;
  facts->SP_land_qstep = land_qstep;
  facts->SP_angle_qstep = dir_qstep;
  facts->SP_dist_qstep_l = dist_qstep_l;
  facts->SP_dist_qstep_r = dist_qstep_r;
  facts->SP_land_qstep_l = land_qstep_l;
  facts->SP_land_qstep_r = land_qstep_r;
  facts->SP_angle_qstep_l = dir_qstep_l;
  facts->SP_angle_qstep_r = dir_qstep_r;
  facts->SP_coach = CoachMode;
  facts->SP_coach_w_referee = CwRMode;
  facts->SP_old_coach_hear = old_hear;
  facts->SP_send_vi_step = sv_st;  
  facts->SP_start_goal_l = start_goal_l;
  facts->SP_start_goal_r = start_goal_r;
  facts->SP_fullstate_l = fullstate_l;
  facts->SP_fullstate_r = fullstate_r;
  facts->SP_drop_time = drop_time; 
  
  lastSensorType = SENSOR_PARAM;
}

void Sensors::doBuildPlayerParam( int player_types,
				  int subs_max,
				  int pt_max,
				  double player_speed_max_delta_min,
				  double player_speed_max_delta_max,
				  double stamina_inc_max_delta_factor,
				  double player_decay_delta_min,
				  double player_decay_delta_max,
				  double inertia_moment_delta_factor,
				  double dash_power_rate_delta_min,
				  double dash_power_rate_delta_max,
				  double player_size_delta_factor,
				  double kickable_margin_delta_min,
				  double kickable_margin_delta_max,
				  double kick_rand_delta_factor,
				  double extra_stamina_delta_min,
				  double extra_stamina_delta_max,
				  double effort_max_delta_factor,
				  double effort_min_delta_factor ) {
  RS_LOG ( LA_SENSORS, "BuildPlayerParam: "
	   << "\nplayer_types: " << player_types
	   << "\nsubs_max: " << subs_max
	   << "\npt_max: " << pt_max
	   << "\nplayer_speed_max_delta_min: " << player_speed_max_delta_min
	   << "\nplayer_speed_max_delta_max: " << player_speed_max_delta_max
	   << "\nstamina_inc_max_delta_factor: " << stamina_inc_max_delta_factor
	   << "\nplayer_decay_delta_min: " << player_decay_delta_min
	   << "\nplayer_decay_delta_max: " << player_decay_delta_max
	   << "\ninertia_moment_delta_factor: " << inertia_moment_delta_factor
	   << "\ndash_power_rate_delta_min: " << dash_power_rate_delta_min
	   << "\ndash_power_rate_delta_max: " << dash_power_rate_delta_max
	   << "\nplayer_size_delta_factor: " << player_size_delta_factor
	   << "\nkickable_margin_delta_min: " << kickable_margin_delta_min
	   << "\nkickable_margin_delta_max: " << kickable_margin_delta_max
	   << "\nkick_rand_delta_factor: " << kick_rand_delta_factor
	   << "\nextra_stamina_delta_min: " << extra_stamina_delta_min
	   << "\nextra_stamina_delta_max: " << extra_stamina_delta_max
	   << "\neffort_max_delta_factor: " << effort_max_delta_factor
	   << "\neffort_min_delta_factor: " << effort_min_delta_factor );
  lastSensorType = SENSOR_PLAYER_PARAM;
}

void Sensors::doBuildPlayerType ( int id,
				  double player_speed_max,
				  double stamina_inc_max,
				  double player_decay,
				  double inertia_moment,
				  double dash_power_rate,
				  double player_size,
				  double kickable_margin,
				  double kick_rand,
				  double extra_stamina,
				  double effort_max,
				  double effort_min ) {
  RS_LOG ( LA_SENSORS, "BuildPlayerType:"
	   << "\nid: " << id
	   << "\nplayer_speed_max: " << player_speed_max
	   << "\nstamina_inc_max: " << stamina_inc_max
	   << "\nplayer_decay: " << player_decay
	   << "\ninertia_moment: " << inertia_moment
	   << "\ndash_power_rate: " << dash_power_rate
	   << "\nplayer_size: " << player_size
	   << "\nkickable_margin: " << kickable_margin
	   << "\nkick_rand: " << kick_rand
	   << "\nextra_stamina: " << extra_stamina
	   << "\neffort_max: " << effort_max
	   << "\neffort_min: " << effort_min );
    lastSensorType = SENSOR_PLAYER_TYPE;
}

void Sensors::doBuildVisual( int time ) {
  RS_LOG ( LA_SENSORS, "BuildVisual: " << time);  
  lastSensorType = SENSOR_VISUAL;
  updateTime(time);
  theSensorInterface->see(parsedLines,
			  parsedMarkers,
			  parsedPlayers,
			  parsedBall);
  parsedLines.clear();
  parsedMarkers.clear();
  parsedPlayers.clear();
  parsedBall = ServerBallObject();
  parsedAgent = ServerAgentObject();
}

SideLineName Sensors::locationToLine() const {
  if (!parsedLocations.empty())
    switch (parsedLocations[0]) {
    case FL_TOP: {
      return SL_TOP;
    }
    case FL_LEFT: {
      return SL_LEFT;
    }
    case FL_RIGHT: {
      return SL_RIGHT;
    }
    case FL_BOTTOM: {
      return SL_BOTTOM;
    }
    default: {
      return SL_UNKNOWN;
    }
    }
  return SL_UNKNOWN;
}
  
void Sensors::doBuildLine( double dist, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildLine: " << dist << ", " << dir);  
  parsedLines.push_back 
    (ServerLineObject (locationToLine(),
		       AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
		       FloatUE(ERROR_QUANTIZE, dist, facts->SP_land_qstep)));
  parsedLocations.clear();
}

void Sensors::doBuildLine( double dir ) {
  RS_LOG ( LA_SENSORS, "BuildLine: " << dir);
  parsedLines.push_back 
    (ServerLineObject (locationToLine(),
		       AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep)));
  parsedLocations.clear();    
}

void Sensors::doBuildTopLocation() {
  RS_LOG ( LA_SENSORS, "BuildTopLocation" );
  parsedLocations.push_back(FL_TOP);
}
  
void Sensors::doBuildBottomLocation() {
  RS_LOG ( LA_SENSORS, "BuildBottomLocation" );
  parsedLocations.push_back(FL_BOTTOM);
}
 
void Sensors::doBuildLeftLocation() {
  RS_LOG ( LA_SENSORS, "BuildLeftLocation" );
  parsedLocations.push_back(FL_LEFT);
}
 
void Sensors::doBuildRightLocation() {
  RS_LOG ( LA_SENSORS, "BuildRightLocation" );
  parsedLocations.push_back(FL_RIGHT);
}
 
void Sensors::doBuildCenterLocation() {
  RS_LOG ( LA_SENSORS, "BuildCenterLocation" );
  parsedLocations.push_back(FL_CENTER);
}
 
void Sensors::doBuildPenaltyLocation() {
  RS_LOG ( LA_SENSORS, "BuildPenaltyLocation" );
  parsedLocations.push_back(FL_PENALTY);
}
 
void Sensors::doBuildGoalLocation() {
  RS_LOG ( LA_SENSORS, "BuildGoalLocation" );
  parsedLocations.push_back(FL_GOAL);
}
 
void Sensors::doBuildFlagOffset( int offset ) {
  RS_LOG ( LA_SENSORS, "BuildFlagOffset" );
  parsedFlagOffset = offset;
}
 
MarkerName Sensors::locationToMarker() const {
  switch (parsedLocations.size()) {
  case 0: {
    return UNKNOWN_MARKER;
  }
  case 1: {
    if (parsedFlagOffset.isKnown()) {
      if (parsedFlagOffset.getValue() == 0) {
	switch (parsedLocations[0]) {
	case FL_LEFT:
	  return FLAG_L0;
	case FL_RIGHT:
	  return FLAG_R0;
	case FL_TOP:
	  return FLAG_T0;
	case FL_BOTTOM:
	  return FLAG_B0;
	default:
	  return UNKNOWN_MARKER;
	}
      }
      else { 
	return UNKNOWN_MARKER;
      }
    }
    else {
      switch (parsedLocations[0]) {
      case FL_CENTER:
	return FLAG_C;      
      case FL_RIGHT:
	return GOAL_R;
      case FL_LEFT:
	return GOAL_L;
      default:
	return UNKNOWN_MARKER;
      } 
    }
  }
  case 2: {
    switch (parsedLocations[0]) {
    case FL_CENTER: {
      switch (parsedLocations[1]) {
      case FL_TOP:
	return FLAG_CT;
      case FL_BOTTOM: 
	return FLAG_CB;
      default:
	return UNKNOWN_MARKER;
      }
    }
    case FL_LEFT: {
      switch (parsedLocations[1]) {      
      case FL_TOP: {
	if (parsedFlagOffset.isKnown()) {
	  switch (parsedFlagOffset.getValue()) {
	  case 10:
	    return FLAG_LT10;
	  case 20: 
	    return FLAG_LT20;
	  case 30:
	    return FLAG_LT30;
	  default:
	    return UNKNOWN_MARKER;
	  }
	}
	return FLAG_LT;
      }
      case FL_BOTTOM: {
	if (parsedFlagOffset.isKnown()) {
	  switch (parsedFlagOffset.getValue()) {
	  case 10: 
	    return FLAG_LB10;
	  case 20:
	    return FLAG_LB20;
	  case 30:
	    return FLAG_LB30;
	  default:
	    return UNKNOWN_MARKER;
	  }
	}
	return FLAG_LB;
      }      
      default:
	return UNKNOWN_MARKER;
      }
    }
    case FL_RIGHT: {
      switch (parsedLocations[1]) {      
      case FL_TOP: {
	if (parsedFlagOffset.isKnown()) {
	  switch (parsedFlagOffset.getValue()) {
	  case 10:
	    return FLAG_RT10;
	  case 20:
	    return FLAG_RT20;
	  case 30:
	    return FLAG_RT30;
	  default:
	    return UNKNOWN_MARKER;
	  }
	}
	return FLAG_LT;
      }
      case FL_BOTTOM: {
	if (parsedFlagOffset.isKnown()) {
	  switch (parsedFlagOffset.getValue()) {
	  case 10:
	    return FLAG_RB10;
	  case 20:
	    return FLAG_RB20;
	  case 30:
	    return FLAG_RB30;
	  default:
	    return UNKNOWN_MARKER;
	  }
	}
	return FLAG_LB;
      }
      default: 
	return UNKNOWN_MARKER;
      }
    }
    case FL_TOP: {
      if (parsedFlagOffset.isKnown()) {
	switch (parsedLocations[1]) {
	case FL_LEFT: {
	  switch (parsedFlagOffset.getValue()) {
	  case 10: 
	    return FLAG_TL10;
	  case 20:
	    return FLAG_TL20;
	  case 30:
	    return FLAG_TL30;
	  case 40:
	    return FLAG_TL40;
	  case 50:
	    return FLAG_TL50;
	  default:
	    return UNKNOWN_MARKER;
	  }
	}
	case FL_RIGHT: {
	  switch (parsedFlagOffset.getValue()) {
	  case 10:
	    return FLAG_TR10;
	  case 20:
	    return FLAG_TR20;
	  case 30:
	    return FLAG_TR30;
	  case 40:
	    return FLAG_TR40;
	  case 50:
	    return FLAG_TR50;
	  default:
	    return UNKNOWN_MARKER;
	  }
	}
	default:
	  return UNKNOWN_MARKER;
	}
      }
      return UNKNOWN_MARKER;
    }
    case FL_BOTTOM: {
      if (parsedFlagOffset.isKnown()) {
	switch (parsedLocations[1]) {
	case FL_LEFT: {
	  switch (parsedFlagOffset.getValue()) {
	  case 10:
	    return FLAG_BL10;
	  case 20:
	    return FLAG_BL20;
	  case 30:
	    return FLAG_BL30;
	  case 40:
	    return FLAG_BL40;
	  case 50:
	    return FLAG_BL50;
	  default:
	    return UNKNOWN_MARKER;
	  }
	}
	case FL_RIGHT: {
	  switch (parsedFlagOffset.getValue()) {
	  case 10:
	    return FLAG_BR10;
	  case 20:
	    return FLAG_BR20;
	  case 30:
	    return FLAG_BR30;
	  case 40:
	    return FLAG_BR40;
	  case 50:
	    return FLAG_BR50;
	  default:
	    return UNKNOWN_MARKER;
	  }
	}       
	default:
	  return UNKNOWN_MARKER;
	}
      }
      return UNKNOWN_MARKER;
    }
    default:
      return UNKNOWN_MARKER;
    }
  }
  case 3: {
    switch (parsedLocations[0]) {
    case FL_PENALTY: {
      switch (parsedLocations[1]) {
      case FL_LEFT: {
	switch (parsedLocations[2]) {
	case FL_TOP:
	  return FLAG_PLT;
	case FL_CENTER:
	  return FLAG_PLC;
	case FL_BOTTOM:
	  return FLAG_PLB;
	default:
	  return UNKNOWN_MARKER;
	}
      }
      case FL_RIGHT: {
	switch (parsedLocations[2]) {
	case FL_TOP:
	  return FLAG_PRT;
	case FL_CENTER:
	  return FLAG_PRC;
	case FL_BOTTOM:
	  return FLAG_PRB;
	default:
	  return UNKNOWN_MARKER;
	}
      }
      default:
	return UNKNOWN_MARKER;
      }
    }
    case FL_GOAL: {
      switch (parsedLocations[1]) {
      case FL_LEFT: {
	switch (parsedLocations[2]) {
	case FL_TOP:
	  return FLAG_GLT;
	case FL_BOTTOM:
	  return FLAG_GLB;
	default:
	  return UNKNOWN_MARKER;
	}
      }
      case FL_RIGHT: {
	switch (parsedLocations[2]) {
	case FL_TOP:
	  return FLAG_GRT;
	case FL_BOTTOM:
	  return FLAG_GRB;
	default:
	  return UNKNOWN_MARKER;
	}
      }
      default:
	return UNKNOWN_MARKER;
      }
    }
    default:
      return UNKNOWN_MARKER;
    }    
  default:
    return UNKNOWN_MARKER;
  }
  }
}

void Sensors::doBuildFlag( bool close, double dist, double dir,
			   double dist_chg, double dir_chg ) {
  RS_LOG ( LA_SENSORS, "BuildFlag: " << close  << ", " << dist
	   << ", " << dir << ", " << dist_chg  << ", " << dir_chg);
  if (!close) {
    parsedMarkers.push_back
      (ServerMarkerObject
       (locationToMarker(),
	close, 
	AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
	FloatUE(ERROR_QUANTIZE, dist, facts->SP_land_qstep),
	FloatUE(ERROR_DISTCHANGE, dist_chg, facts->SP_dist_qstep,
		dist, facts->CP_dist_change_qstep),
	AngleDegUE(ERROR_QUANTIZE, dir_chg, facts->CP_dir_change_qstep)));
  parsedLocations.clear();
  parsedFlagOffset = UIntUnknown();
  }
}

void Sensors::doBuildFlag( bool close, double dist, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildFlag: " << close << ", " << dist
	   << ", " << dir);
  if (!close) {
    parsedMarkers.push_back
      (ServerMarkerObject
       (locationToMarker(),
	close,
	AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
	FloatUE(ERROR_QUANTIZE, dist, facts->SP_land_qstep)));
    parsedLocations.clear();
    parsedFlagOffset = UIntUnknown();
  }
}
 
void Sensors::doBuildFlag( bool close, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildFlag: " << close  << ", " << dir);
  if (!close) {
    parsedMarkers.push_back
      (ServerMarkerObject
       (locationToMarker(), 
	close,
	AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep)));
    parsedLocations.clear();
    parsedFlagOffset = UIntUnknown();
  }
}

void Sensors::doBuildGoal( bool close,
			   double dist, double dir,
			   double dist_chg, double dir_chg ) {
  RS_LOG ( LA_SENSORS, "BuildGoal:" << close  << ", " << dist
	   << ", " << dir  << ", " << dist_chg  << ", " << dir_chg);
  if (!close)
    doBuildFlag( close, dist, dir, dist_chg, dir_chg);
}

void Sensors::doBuildGoal( bool close, double dist, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildGoal: " << close  << ", " << dist
	   << ", " << dir);
  if (!close)
    doBuildFlag( close, dist, dir);
}

void Sensors::doBuildGoal( bool close, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildGoal: " << close  << ", " << dir );
  if (!close)
    doBuildFlag( close, dir);
}

void Sensors::doBuildBall( bool close,
			   double dist, double dir,
			   double dist_chg, double dir_chg ) {
  RS_LOG ( LA_SENSORS, "BuildBall: " << close << ", " << dist
	   << ", " << dir << ", " << dist_chg << ", " << dir_chg);
  parsedBall = 
    ServerBallObject
    (close,
     AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
     FloatUE(ERROR_QUANTIZE, dist, facts->SP_dist_qstep),
     FloatUE(ERROR_DISTCHANGE, dist_chg, facts->SP_dist_qstep,
	     dist, facts->CP_dist_change_qstep),
     AngleDegUE(ERROR_QUANTIZE, dir_chg, facts->CP_dir_change_qstep));
}

void Sensors::doBuildBall( bool close, double dist, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildBall: " << close  << ", " << dist
	   << ", " << dir);
  parsedBall =
    ServerBallObject(close,
		     AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
		     FloatUE(ERROR_QUANTIZE, dist, facts->SP_dist_qstep));    
}

void Sensors::doBuildBall( bool close, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildBall:"  << ", " << close << ", " << dir);
  parsedBall =
    ServerBallObject(close,
		     AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep));
}

void Sensors::doBuildPlayer( bool close,
			     double dist, double dir,
			     double dist_chg, double dir_chg,
			     double orientation, double head_orientation ) {
  RS_LOG ( LA_SENSORS, "BuildPlayer: " << close  << ", " << dist
	   << ", " << dir << ", " << dist_chg << ", " << dir_chg
	   << ", " << orientation << ", " << head_orientation);
  parsedPlayers.push_back
    (ServerPlayerObject
     (close,
      AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
      parsedGoalie,
      parsedTeam,
      parsedUniformNumber,
      FloatUE(ERROR_QUANTIZE, dist, facts->SP_dist_qstep),
      FloatUE(ERROR_DISTCHANGE, dist_chg, facts->SP_dist_qstep,
	      dist, facts->CP_dist_change_qstep),
      AngleDegUE(ERROR_QUANTIZE, dir_chg, facts->CP_dir_change_qstep),
      AngleDegUE(ERROR_QUANTIZE, orientation, facts->SP_angle_qstep),
      AngleDegUE(ERROR_QUANTIZE, head_orientation, facts->SP_angle_qstep)));
  parsedTeam = T_UNKNOWN;
  parsedUniformNumber = UIntUnknown();
  parsedGoalie = false;  
}

void Sensors::doBuildPlayer( bool close,
			     double dist, double dir,
			     double dist_chg, double dir_chg ) {
  RS_LOG ( LA_SENSORS, "BuildPlayer: " << close << ", " << dist
	   << ", " << dir << ", " << dist_chg << ", " << dir_chg);
  parsedPlayers.push_back
    (ServerPlayerObject
     (close,
      AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
      parsedGoalie,
      parsedTeam,
      parsedUniformNumber,
      FloatUE(ERROR_QUANTIZE, dist, facts->SP_dist_qstep),
      FloatUE(ERROR_DISTCHANGE, dist_chg, facts->SP_dist_qstep,
	      dist, facts->CP_dist_change_qstep),
      AngleDegUE(ERROR_QUANTIZE, dir_chg, facts->CP_dir_change_qstep)));
  parsedTeam = T_UNKNOWN;
  parsedUniformNumber = UIntUnknown();
  parsedGoalie = false;
}

void Sensors::doBuildPlayer( bool close, double dist, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildPlayer: " << close  << ", " << dist
	   << ", " << dir);
  parsedPlayers.push_back
    (ServerPlayerObject
     (close,
      AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
      parsedGoalie,
      parsedTeam,
      parsedUniformNumber,
      FloatUE(ERROR_QUANTIZE, dist, facts->SP_dist_qstep)));
  parsedTeam = T_UNKNOWN;
  parsedUniformNumber = UIntUnknown();
  parsedGoalie = false;
}
 
void Sensors::doBuildPlayer( bool close, double dir ) {
  RS_LOG ( LA_SENSORS, "BuildPlayer:" << close  << ", " << dir);
  parsedPlayers.push_back
    (ServerPlayerObject
     (close,
      AngleDegUE(ERROR_QUANTIZE, dir, facts->SP_angle_qstep),
      parsedGoalie,
      parsedTeam,
      parsedUniformNumber));
  parsedTeam = T_UNKNOWN;
  parsedUniformNumber = UIntUnknown();
  parsedGoalie = false;
}
   
void Sensors::doBuildTeamName( const std::string& name) {
  RS_LOG ( LA_SENSORS, "BuildTeamName: " << name);
  if (name == facts->CP_team_name)
    parsedTeam = T_OUR_TEAM;
  else
    parsedTeam = T_THEIR_TEAM;
}

void Sensors::doBuildUNum( int unum ) {
  RS_LOG ( LA_SENSORS, "BuildUNum: " << unum);
  parsedUniformNumber = unum;
}

void Sensors::doBuildGoalie() {
  RS_LOG ( LA_SENSORS, "BuildGoalie" );
  parsedGoalie = true;
}

void Sensors::doBuildScore( int time, int our, int opp ) {
  RS_LOG ( LA_SENSORS, "BuildScore: " << time  << ", " << our << ", " << opp);
  lastSensorType = SENSOR_AURAL_SCORE;
  updateTime(time);
  doBuildScore(our, opp);
}

void Sensors::doBuildInit( int unum ) {
  RS_LOG ( LA_SENSORS, "BuildInit: " << unum);
  theSensorInterface->hearReferee(parsedPlayMode);
  std::string agentSide;
  if (parsedSide == S_LEFT) {
    facts->setSides(S_LEFT, S_RIGHT);
    agentSide = "left";
  }
  else {
    facts->setSides(S_RIGHT, S_LEFT);
    agentSide = "right";
  }
  facts->setUniformNumber(unum);
  std::cout << "-----------------------------------------------------------"
	    << "-------------" << std::endl
	    << "Connected player " << unum << " on team '"
	    << facts->CP_team_name << "'" << std::endl
	    << "He's playing on the " << agentSide << " side of the field."
	    << std::endl << "----------------------------------------------"
	    << "--------------------------" << std::endl;

  parsedSide = S_UNKNOWN;
  parsedPlayMode = PM_UNKNOWN;  
  lastSensorType = SENSOR_INIT;
}
    
void Sensors::doBuildCoachAudio( int time, const std::string& msg ) {
  RS_LOG ( LA_SENSORS, "BuildCoachAudio: " << time << ", " << msg);
  lastSensorType = SENSOR_AURAL_COACH;
  updateTime(time);
  theSensorInterface->hearCoach(msg);    
}


void Sensors::doBuildRefAudio( int time ) {
  RS_LOG ( LA_SENSORS, "BuildRefAudio: " << time );
  lastSensorType = SENSOR_AURAL_REFEREE;
  updateTime(time);
  theSensorInterface->hearReferee(parsedPlayMode);  
  parsedPlayMode = PM_UNKNOWN;
}


void Sensors::doBuildGoalRefAudio( int time, int score ) {
  RS_LOG ( LA_SENSORS, "BuildGoalRefAudio: " << time << ", " << score);
  lastSensorType = SENSOR_AURAL_REFEREE;
  updateTime(time);
  facts->setScore(parsedSide, score);
  if (parsedSide == facts->getOurSide())
    theSensorInterface->hearReferee(PM_OUR_GOAL);
  else
    theSensorInterface->hearReferee(PM_THEIR_GOAL);
  parsedSide = S_UNKNOWN;
}

void Sensors::doBuildUnknownRefAudio( int time, const std::string& message ) {
  RS_LOG ( LA_SENSORS, "BuildUnknownRefAudio: " << time << ", " << message);  
  RS_LOG ( LA_WARNING, "BuildUnknownRefAudio - not implemented " 
	   << time << ", " << message);
  lastSensorType = SENSOR_AURAL_REFEREE;
  updateTime(time);
}

void Sensors::doBuildPlayerAudio( int time, double dir, 
				  const std::string& msg ) {
  RS_LOG ( LA_SENSORS, "BuildPlayerAudio: " << time << ", " 
	   << dir << ", " << msg);
  lastSensorType = SENSOR_AURAL_PLAYER;
  updateTime(time);
  theSensorInterface->hearPlayer(AngleDegUE(ERROR_QUANTIZE,
					    dir,
					    facts->SP_angle_qstep),
				 msg);
}

void Sensors::doBuildSelfAudio( int time, const std::string& msg ) {
  RS_LOG ( LA_SENSORS, "BuildSelfAudio: " << time << ", " << msg);
  lastSensorType = SENSOR_AURAL_SELF;
  updateTime(time);
  theSensorInterface->hearSelf(msg);  
}

void Sensors::doBuildSubstitution( int unum, int type ) {
  RS_LOG ( LA_SENSORS, "BuildSubstitution: " << unum << ", " << type);
  // Todo!
  lastSensorType = SENSOR_SUBSTITUTION;
}
    
void Sensors::doBuildSubstitution( int unum ) {
  RS_LOG ( LA_SENSORS, "BuildSubstitution: " << unum);
  // Todo!
  lastSensorType = SENSOR_SUBSTITUTION;
}

void Sensors::doBuildCantReconnect() {
  RS_LOG ( LA_SENSORS, "BuildCantReconnect" );
  RS_WARNING ( "BuildCantReconnect" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildInitError() {
  RS_LOG ( LA_SENSORS, "BuildInitError" );
  RS_WARNING(  "BuildInitError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildNoMoreTeamOrPlayerOrGoalieError() {
  RS_LOG ( LA_SENSORS, "BuildNoMoreTeamOrPlayerOrGoalieError" );
  RS_WARNING ( "BuildNoMoreTeamOrPlayerOrGoalieError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildNoSuchTeamOrAlreadyHaveCoachError() {
  RS_LOG ( LA_SENSORS, "BuildNoSuchTeamOrAlreadyHaveCoachError" );
  RS_WARNING ( "BuildNoSuchTeamOrAlreadyHaveCoachError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildTooManyMovesError() {
  RS_LOG ( LA_SENSORS, "BuildTooManyMovesError" );
  RS_WARNING ( "BuildTooManyMovesError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildUnknownCommandError() {
  RS_LOG ( LA_SENSORS, "BuildUnknownCommandError" );
  RS_WARNING ( "BuildUnknownCommandError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildIllegalCommandError() {
  RS_LOG ( LA_SENSORS, "BuildIllegalCommandError" );
}

void Sensors::doBuildSayMessageTooLongError() {
  RS_LOG ( LA_SENSORS, "BuildSayMessageTooLongError" );
  RS_WARNING ( "BuildSayMessageTooLongError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildIllegalModeError() {
  RS_LOG ( LA_SENSORS, "BuildIllegalModeError" );
  RS_WARNING ( "BuildIllegalModeError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildIllegalObjectFormError() {
  RS_LOG ( LA_SENSORS, "BuildIllegalObjectFormError" );
  RS_WARNING ( "BuildIllegalObjectFormError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildSaidTooManyFreeformMessagesError() {
  RS_LOG ( LA_SENSORS, "BuildSaidTooManyFreeformMessagesError" );
  RS_WARNING ( "BuildSaidTooManyFreeformMessagesError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildCannotSayFreeformWhilePlayonError() {
  RS_LOG ( LA_SENSORS, "BuildCannotSayFreeformWhilePlayonError" );
  RS_WARNING ( "BuildCannotSayFreeformWhilePlayonError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildSaidTooManyMetaMessagesError() {
  RS_LOG ( LA_SENSORS, "BuildSaidTooManyMetaMessagesError" );
  RS_WARNING ( "BuildSaidTooManyMetaMessagesError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildSaidTooManyAdviceMessagesError() {
  RS_LOG ( LA_SENSORS, "BuildSaidTooManyAdviceMessagesError" );
  RS_WARNING ( "BuildSaidTooManyAdviceMessagesError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildSaidTooManyInfoMessagesError() {
  RS_LOG ( LA_SENSORS, "BuildSaidTooManyInfoMessagesError" );
  RS_WARNING ( "BuildSaidTooManyInfoMessagesError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildSaidTooManyDefineMessagesError() {
  RS_LOG ( LA_SENSORS, "BuildSaidTooManyDefineMessagesError" );
  RS_WARNING ( "BuildSaidTooManyDefineMessagesError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildCouldNotParseSayError() {
  RS_LOG ( LA_SENSORS, "BuildCouldNotParseSayError" );
  RS_WARNING ( "BuildCouldNotParseSayError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildSaidTooManyMessagesError() {
  RS_LOG ( LA_SENSORS, "BuildSaidTooManyMessagesError" );
  RS_WARNING ( "BuildSaidTooManyMessagesError" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildUnknownError( const std::string& error ) {
  RS_LOG ( LA_SENSORS, "BuildUnknownError: " << error);
  RS_WARNING ( "BuildUnknownError: " << error);
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildNoTeamFoundWarning() {
  RS_LOG ( LA_SENSORS, "BuildNoTeamFoundWarning" );
  RS_WARNING ( "BuildNoTeamFoundWarning" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildNoSuchPlayerWarning() {
  RS_LOG ( LA_SENSORS, "BuildNoSuchPlayerWarning" );
  RS_WARNING ( "BuildNoSuchPlayerWarning" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildCannotSubWhilePlayOnWarning() {
  RS_LOG ( LA_SENSORS, "BuildCannotSubWhilePlayOnWarning" );
  RS_WARNING ( "BuildCannotSubWhilePlayOnWarning" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildNoSubsLeftWarning() {
  RS_LOG ( LA_SENSORS, "BuildNoSubsLeftWarning" );
  RS_WARNING ( "BuildNoSubsLeftWarning" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildMaxOfThatPlayerTypeOnFieldWarning() {
  RS_LOG ( LA_SENSORS, "BuildMaxOfThatPlayerTypeOnFieldWarning" );
  RS_WARNING ( "BuildMaxOfThatPlayerTypeOnFieldWarning" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildCannotChangeGoalieWarning() {
  RS_LOG ( LA_SENSORS, "BuildCannotChangeGoalieWarning" );
  RS_WARNING ( "BuildCannotChangeGoalieWarning" );
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildUnknownWarning(const std::string& warning) {
  RS_LOG ( LA_SENSORS, "BuildUnknownWarning: " << warning);
  RS_WARNING ( "BuildUnknownWarning: " << warning);
  lastSensorType = SENSOR_ERROR;
}

void Sensors::doBuildClangVerOK( int min, int max ) {
  RS_LOG ( LA_SENSORS, "BuildClangVerOK: " << min << ", " << max );
}

void Sensors::doBuildUnknownOK(const std::string& msg) {
  RS_LOG ( LA_SENSORS, "BuildUnknownOK: " << msg);
}
          
void Sensors::doBuildFoulLeftPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulLeftPlayMode: ");
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FOUL;  
}

void Sensors::doBuildFoulRightPlayMode() {
  RS_LOG ( LA_SENSORS, "BuildFoulRightPlayMode: ");
  if (facts->getOurSide() == S_LEFT)
    parsedPlayMode = PM_OUR_FOUL;
  else if (facts->getOurSide() == S_RIGHT)
    parsedPlayMode = PM_THEIR_FOUL;
}

void Sensors::doBuildGlobalVisual(int time) {
  RS_LOG ( LA_SENSORS, "BuildGlobalVisual: " << time);
  lastSensorType = SENSOR_COACH;
  updateTime(time);
  theSensorInterface->seeGlobal(parsedCoachPlayers,
				parsedCoachBall);
  parsedCoachPlayers.clear();
  parsedCoachBall = CoachBallObject();  
}

void Sensors::doBuildGlobalGoal(double x, double y) {
  RS_LOG ( LA_SENSORS, "BuildGlobalGoal: " << x << ", " << y);
}

void Sensors::doBuildGlobalBall(double x, double y, double dx, double dy) {
  RS_LOG ( LA_SENSORS, "BuildGlobalBall: " << x << ", " << y 
	   << ", " << dx << ", " << dy);
  parsedCoachBall = CoachBallObject(x, y, dx, dy);
}

void Sensors::doBuildGlobalPlayer(double x, double y, double dx, double dy, 
				  double bodydir, double neckdir) {
  RS_LOG ( LA_SENSORS, "BuildGlobalPlayer: " << x << ", " << y 
	   << ", " << dx << ", " << dy << ", " << bodydir << ", " 
	   << neckdir);
  parsedCoachPlayers.push_back
    (CoachPlayerObject
     (x, y, dx, dy, 
      parsedTeam,
      parsedUniformNumber.getValue(),
      parsedGoalie,
      bodydir, 
      neckdir));
  parsedTeam = T_UNKNOWN;
  parsedUniformNumber = UIntUnknown();
  parsedGoalie = false;
}

void Sensors::doBuildCoachInit() {
  RS_LOG ( LA_SENSORS, "BuildCoachInit");
  std::string coachSide;
  if (parsedSide == S_LEFT) {    
    facts->setSides(S_LEFT, S_RIGHT);
    coachSide = "left";
  }
  else {
    facts->setSides(S_RIGHT, S_LEFT);
    coachSide = "right";
  }
  std::cout << "-----------------------------------------------------------"
	    << "-------------" << std::endl
	    << "Connected coach on team '" << facts->CP_team_name
	    << "'" << std::endl
	    << "He's coaching the team on the " << coachSide << " side of the field."
	    << std::endl << "----------------------------------------------"
	    << "--------------------------" << std::endl;
  parsedSide = S_UNKNOWN;
  parsedPlayMode = PM_UNKNOWN;  
  lastSensorType = SENSOR_INIT;
}

void Sensors::doBuildPlayerAudio(int unum, const std::string& msg) {
  RS_LOG ( LA_SENSORS, "BuildPlayerAudio: " << unum << ", '"
	   << msg << "'");
  theSensorInterface->hearPlayer(unum, msg);
  lastSensorType = SENSOR_AURAL_PLAYER;  
}

void Sensors::doBuildEyeOK(bool ok) {
  RS_LOG (LA_SENSORS, "BuildEyeOK: " << ok);
  lastSensorType = SENSOR_OK;
}

void Sensors::doBuildBall( double x, double y, 
			   double delta_x, double delta_y ) {
  RS_LOG ( LA_SENSORS, "BuildBall: " << x << ", " << y << ", "
	   << delta_x << ", " << delta_y);
  RS_WARNING("This doBuildBall is not supported!");
  lastSensorType = SENSOR_UNKNOWN;
}

void Sensors::doBuildPlayer( int unum, double x, double y, 
			     double delta_x, double delta_y,
			     double orientation, double head_orientation,
			     int stamina, double effort, double recovery ) {
  RS_LOG ( LA_SENSORS, "BuildPlayer: " << unum << ", " << x  << ", " 
	   << y  << ", " << delta_x  << ", " << delta_y  << ", " << orientation
	   << ", " << head_orientation << ", " << stamina  << ", " << effort
	   << ", " << recovery);
  RS_WARNING("This doBuildPlayer is not supported!");
  lastSensorType = SENSOR_UNKNOWN;
}

void Sensors::doBuildInit() {
  RS_LOG ( LA_SENSORS, "BuildInit" );
  RS_WARNING("This doBuildInit is not supported!");
  lastSensorType = SENSOR_UNKNOWN;
}

RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Sensors.icc"
#endif
