/* -*- Mode: C++ -*-
 * $Id: Sensors.h,v 1.16 2002/09/26 15:23:41 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Sensors.h
 * Unitname: BasicSystem
 * $Revision: 1.16 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: s02davro $ $Date: 2002/09/26 15:23:41 $
 *
 *
 * HISTORY
 *
 * $Log: Sensors.h,v $
 * Revision 1.16  2002/09/26 15:23:41  s02davro
 * Bugfixes, close, team, coach
 *
 * Revision 1.15  2002/09/23 10:28:42  s02davro
 * Find bug in (close), this information needs to be handled - todo
 *
 * Revision 1.14  2002/09/21 15:38:58  s02davro
 * Update to new (1.0.3) RCCParser
 *
 * Revision 1.13  2002/08/31 15:06:16  s02davro
 * Redesign (again) now handles the coach fullstate information messages
 *
 * Revision 1.12  2002/08/02 21:51:50  s02davro
 * Complete rewrite, now uses rccparser-1.0 and clangparser from the soccerserver.
 *
 * Revision 1.11  2002/01/15 17:19:41  frehe
 * Improved the coach parser.
 *
 * Revision 1.10  2001/11/15 16:48:48  frehe
 * Added support for a coach.
 *
 * Revision 1.9  2001/09/17 21:54:00  frehe
 * Readded support for a debug coach.
 *
 * Revision 1.8  2001/09/17 21:30:09  andbo928
 * Implemented an ugly workaround to get ballspeeds working.
 *
 * Revision 1.7  2001/08/19 23:32:21  frehe
 * Updated 2001-08-19
 *
 * Revision 1.6  2001/08/17 07:15:13  frehe
 * Made changes to compile RoboSoc with RCCParser-0.10.
 *
 * Revision 1.5  2001/08/13 16:43:14  frehe
 * Made updates to compile with new autotools and RCCParser.
 *
 * Revision 1.4  2001/08/13 07:33:16  frehe
 * Updated from laptop 2001-08-13.
 *
 * Revision 1.3  2001/08/02 17:04:22  frehe
 * Updated 2001-08-02
 *
 * Revision 1.2  2000/09/04 08:46:45  frehe
 * Fixed a naming error in the sensors.
 *
 * Revision 1.1  2000/09/03 20:41:10  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Sensors.h
 * Include file for the class Sensors.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.16 $
 * $Author: s02davro $
 * $Date: 2002/09/26 15:23:41 $
 */


#ifndef SENSORS_H
#define SENSORS_H


#include <queue>

#include <rccparser/rccparser.h>

#include "basic_system_types.h"
#include "ServerCommand.h"
#include "WorldFacts.h"

RS_BEGIN_NAMESPACE


class Sensors;
class Controller;
class SensorInterface;

/**
 * @class Sensors Sensors.h
 * This class is responsible for parsing the server messages
 * and storing them in the sensor interface for further use
 * by the agent.
 *
 * @todo Handle new serverinformation (tackle, point)
 * @todo Syncronization in Controller may force changes here.
 * @todo Handle the information about close objects.
 * @todo Analyze feedback in a correct way.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.16 $
 * $Author: s02davro $
 * $Date: 2002/09/26 15:23:41 $
 */

class Sensors : public rcc::Parser {
public:
  /** Constructor, needs a SensorInterface pointer and a coachlanguageparser. 
   *  The rcss::Parser is needed for the rcc:Parser to be able to parse
   *  coach messages. The sensorInterface is called when we see, hear or
   *  sense something new. */
  Sensors(SensorInterface*, 
	  rcss::Parser& coachMessageParser);
  
  /** Init the Sensors
   *  We need a pointer to the Controller to handle holes and clashes 
   *  (Is our commands getting to the soccerserver in time. */
  void init(Controller*);

  /** Analyze new sensor data.
   *  The data is given to the rcc::Parser which calls our doBuildfunctions. */
  void analyze(char* sensor_data);

  /** Analyze feedback from the actuators. */
  void analyze(ServerCommand*);  

  /** Get the current time. 
   *  This meens the last known server time, according to the sensor
   *  information. */
  UInt getCurrentTime() const throw(); 

  /** Get the type of the last sensor information.
   *  This is used by the sensorinterface call the right functions. */
  SensorType getLastSensorType() const throw();

protected:
  /** Set the type of the last sensor information. 
   *  This is beeing done in the doBuild functions called by the parser. */
  void setLastSensorType(SensorType sensor_type) throw();

  /** Set the current (last-known-server) time. */
  void setCurrentTime(UInt time) throw();

  /** Update the time in the sensors and the controller */
  /** This according to the time information in the sensor data. */
  void updateTime(int time);

  /** @name Builder functions.
   *  This functions are called from the rcc:Parser during the parsing of
   *  sensor data. The functions stores the results in local objects 
   *  (parsed*) and sends the chuncked information to SensorInterface for
   *  more analyzation. */
  //@{
  void doMsgParsed();
  void doParsedCLang( const std::string& msg );
  void doBuildFullState( int time ); 
  void doBuildCatchBallLeftPlayMode();
  void doBuildCatchBallRightPlayMode();
  void doBuildBeforeKickOffPlayMode();
  void doBuildTimeOverPlayMode();
  void doBuildPlayOnPlayMode();
  void doBuildKickOffLeftPlayMode();
  void doBuildKickOffRightPlayMode(); 
  void doBuildKickInLeftPlayMode(); 
  void doBuildKickInRightPlayMode();
  void doBuildFreeKickLeftPlayMode();
  void doBuildFreeKickRightPlayMode();
  void doBuildCornerKickLeftPlayMode();
  void doBuildCornerKickRightPlayMode(); 
  void doBuildGoalKickLeftPlayMode();
  void doBuildGoalKickRightPlayMode(); 
  void doBuildAfterGoalLeftPlayMode(); 
  void doBuildAfterGoalRightPlayMode(); 
  void doBuildDropBallPlayMode(); 
  void doBuildOffSideLeftPlayMode();
  void doBuildOffsideRightPlayMode(); 
  void doBuildPenaltyKickLeftPlayMode(); 
  void doBuildPenaltyKickRightPlayMode(); 
  void doBuildFirstHalfOverPlayMode(); 
  void doBuildPausePlayMode(); 
  void doBuildHumanPlayMode(); 
  void doBuildFoulChargeLeftPlayMode();
  void doBuildFoulChargeRightPlayMode();
  void doBuildFoulPushLeftPlayMode();
  void doBuildFoulPushRightPlayMode();
  void doBuildFoulMultipleAttackerLeftPlayMode();
  void doBuildFoulMultipleAttackerRightPlayMode();
  void doBuildFoulBallOutLeftPlayMode();
  void doBuildFoulBallOutRightPlayMode();
  void doBuildBackPassLeftPlayMode();
  void doBuildBackPassRightPlayMode();
  void doBuildFreeKickFaultLeftPlayMode();
  void doBuildFreeKickFaultRightPlayMode();
  void doBuildScore( int our, int opp );
  void doBuildBall( double x, double y, 
		    double delta_x, double delta_y );
  void doBuildPlayer( int unum,
		      double x, double y, 
		      double delta_x, double delta_y,
		      double orientation, double head_orientation,
		      int stamina, double effort, double recovery );
  void doBuildLeftSide();
  void doBuildRightSide();
  void doBuildOppSide();
  void doBuildOurSide();
  void doBuildNeutralSide();
  void doBuildSenseBody( int time, 
			 double stamina,
			 double effort,
			 double speed_mag,
			 double speed_head,
			 double head_angle,
			 int kick_count,
			 int dash_count,
			 int turn_count,
			 int say_count,
			 int turn_neck_count,
			 int catch_count,
			 int move_count,
			 int chg_view_count );
  void doBuildViewMode();
  void doBuildViewQualityHigh();
  void doBuildViewQualityLow();
  void doBuildViewWidthNarrow();
  void doBuildViewWidthNormal();
  void doBuildViewWidthWide();
  void doBuildServerParam ( double gwidth, 
			    double inertia_moment, 
			    double psize, 
			    double pdecay, 
			    double prand,
			    double pweight, 
			    double pspeed_max, 
			    double paccel_max, 
			    double stamina_max, 
			    double stamina_inc,
			    double recover_init, 
			    double recover_dthr, 
			    double recover_min, 
			    double recover_dec, 
			    double effort_init,
			    double effort_dthr, 
			    double effort_min, 
			    double effort_dec, 
			    double effort_ithr, 
			    double effort_inc,
			    double kick_rand, 
			    int team_actuator_noise, 
			    double prand_factor_l, 
			    double prand_factor_r, 
			    double kick_rand_factor_l,
			    double kick_rand_factor_r, 
			    double bsize, 
			    double bdecay, 
			    double brand, 
			    double bweight,
			    double bspeed_max, 
			    double baccel_max, 
			    double dprate, 
			    double kprate, 
			    double kmargin,
			    double ctlradius, 
			    double ctlradius_width, 
			    double maxp, 
			    double minp, 
			    double maxm,
			    double minm, 
			    double maxnm, 
			    double minnm, 
			    double maxn, 
			    double minn,
			    double visangle, 
			    double visdist, 
			    double windir, 
			    double winforce, 
			    double winang,
			    double winrand, 
			    double kickable_area, 
			    double catch_area_l, 
			    double catch_area_w, 
			    double catch_prob,
			    int goalie_max_moves, 
			    double ckmargin, 
			    double offside_area, 
			    int win_no, 
			    int win_random,
			    int say_cnt_max, 
			    int SayCoachMsgSize, 
			    int clang_win_size, 
			    int clang_define_win, 
			    int clang_meta_win,
			    int clang_advice_win, 
			    int clang_info_win, 
			    int clang_mess_delay, 
			    int clang_mess_per_cycle, 
			    int half_time,
			    int sim_st, 
			    int send_st, 
			    int recv_st, 
			    int sb_step, 
			    int lcm_st,
			    int SayMsgSize, 
			    int hear_max, 
			    int hear_inc, 
			    int hear_decay, 
			    int cban_cycle,
			    int slow_down_factor, 
			    int useoffside, 
			    int kickoffoffside, 
			    double offside_kick_margin, 
			    double audio_dist,
			    double dist_qstep, 
			    double land_qstep, 
			    double dir_qstep, 
			    double dist_qstep_l, 
			    double dist_qstep_r,
			    double land_qstep_l, 
			    double land_qstep_r, 
			    double dir_qstep_l, 
			    double dir_qstep_r, 
			    int CoachMode,
			    int CwRMode, 
			    int old_hear, 
			    int sv_st, 
			    int start_goal_l, 
			    int start_goal_r,
			    int fullstate_l, 
			    int fullstate_r, 
			    int drop_time );
  void doBuildPlayerParam( int player_types,
			   int subs_max,
			   int pt_max,
			   double player_speed_max_delta_min,
			   double player_speed_max_delta_max,
			   double stamina_inc_max_delta_factor,
			   double player_decay_delta_min,
			   double player_decay_delta_max,
			   double inertia_moment_delta_factor,
			   double dash_power_rate_delta_min,
			   double dash_power_rate_delta_max,
			   double player_size_delta_factor,
			   double kickable_margin_delta_min,
			   double kickable_margin_delta_max,
			   double kick_rand_delta_factor,
			   double extra_stamina_delta_min,
			   double extra_stamina_delta_max,
			   double effort_max_delta_factor,
			   double effort_min_delta_factor );
  void doBuildPlayerType ( int id,
			   double player_speed_max,
			   double stamina_inc_max,
			   double player_decay,
			   double inertia_moment,
			   double dash_power_rate,
			   double player_size,
			   double kickable_margin,
			   double kick_rand,
			   double extra_stamina,
			   double effort_max,
			   double effort_min );
  void doBuildVisual( int time );
  void doBuildLine( double dist, double dir );
  void doBuildLine( double dir );
  void doBuildTopLocation(); 
  void doBuildBottomLocation(); 
  void doBuildLeftLocation(); 
  void doBuildRightLocation(); 
  void doBuildCenterLocation(); 
  void doBuildPenaltyLocation(); 
  void doBuildGoalLocation();
  void doBuildFlagOffset( int offset );
  void doBuildFlag( bool close,
		    double dist, double dir,
		    double dist_chg, double dir_chg );
  void doBuildFlag( bool close, double dist, double dir );
  void doBuildFlag( bool close, double dir );
  void doBuildGoal( bool close,
		    double dist, double dir,
		    double dist_chg, double dir_chg );
  void doBuildGoal( bool close, double dist, double dir );
  void doBuildGoal( bool close, double dir );
  void doBuildBall( bool close,
		    double dist, double dir,
		    double dist_chg, double dir_chg );
  void doBuildBall( bool close, double dist, double dir );
  void doBuildBall( bool close, double dir );
  void doBuildPlayer( bool close,
		      double dist, double dir,
		      double dist_chg, double dir_chg,
		      double orientation, double head_orientation );
  void doBuildPlayer( bool close,
		      double dist, double dir,
		      double dist_chg, double dir_chg );
  void doBuildPlayer( bool close, double dist, double dir );
  void doBuildPlayer( bool close, double dir );    
  void doBuildTeamName( const std::string& name );
  void doBuildUNum( int unum );
  void doBuildGoalie();
  void doBuildScore( int time, int our, int opp );
  void doBuildInit( int unum );
  void doBuildInit();    
  void doBuildCoachAudio( int time, const std::string& msg );
  void doBuildRefAudio( int time );
  void doBuildGoalRefAudio( int time, int score );
  void doBuildUnknownRefAudio( int time, const std::string& message );
  void doBuildPlayerAudio( int time, double dir, const std::string& msg );
  void doBuildSelfAudio( int time, const std::string& msg );
  void doBuildSubstitution( int unum, int type );    
  void doBuildSubstitution( int unum );
  void doBuildCantReconnect();
  void doBuildInitError();
  void doBuildNoMoreTeamOrPlayerOrGoalieError();    
  void doBuildNoSuchTeamOrAlreadyHaveCoachError(); 
  void doBuildTooManyMovesError();
  void doBuildUnknownCommandError();
  void doBuildIllegalCommandError();
  void doBuildSayMessageTooLongError(); 
  void doBuildIllegalModeError(); 
  void doBuildIllegalObjectFormError(); 
  void doBuildSaidTooManyFreeformMessagesError();
  void doBuildCannotSayFreeformWhilePlayonError();
  void doBuildSaidTooManyMetaMessagesError();
  void doBuildSaidTooManyAdviceMessagesError();
  void doBuildSaidTooManyInfoMessagesError();
  void doBuildSaidTooManyDefineMessagesError();
  void doBuildCouldNotParseSayError();
  void doBuildSaidTooManyMessagesError();
  void doBuildUnknownError( const std::string& error );
  void doBuildNoTeamFoundWarning();
  void doBuildNoSuchPlayerWarning();
  void doBuildCannotSubWhilePlayOnWarning();
  void doBuildNoSubsLeftWarning();
  void doBuildMaxOfThatPlayerTypeOnFieldWarning();
  void doBuildCannotChangeGoalieWarning();
  void doBuildUnknownWarning( const std::string& warning );    
  void doBuildClangVerOK( int min, int max );          
  void doBuildUnknownOK( const std::string& msg ); 
  void doBuildFoulLeftPlayMode();
  void doBuildFoulRightPlayMode();
  void doBuildGlobalVisual(int time);
  void doBuildGlobalGoal(double x, double y);
  void doBuildGlobalBall(double x, double y, double dx, double dy);
  void doBuildGlobalPlayer(double x, double y, double dx, double dy, 
			   double stamina, double recovery);
  void doBuildCoachInit();
  void doBuildPlayerAudio(int unum, const std::string& msg);
  void doBuildEyeOK(bool ok);
  //@}
  
  /** @name Help fuctions for converting Locations to Lines and Markers. */
  //@{
  /** Convert the locations to a line. */
  SideLineName locationToLine() const;

  /** Convert the locations to a markerer. */
  MarkerName locationToMarker() const;
  //@}

  /** The facts about the world. */
  WorldFacts* facts;

  /** Pointer to the agents controller. */
  Controller* theController;

  /** Pointer to the agents sensorInterface. */
  SensorInterface* theSensorInterface;

  /** The type of the last sensor information. 
   *  Set to SENSOR_UNKNOWN during the parsing of new sensor information.
   *  See basic_system_types.h for more information. */
  SensorType lastSensorType;

  /** The current (last-known-server) time. */
  UInt currentTime;

  /** A queue of commands the agent will execute. */
  std::queue<CommandType> commandQueue;
      
  /** Hidden copy-constructor. */
  Sensors(const Sensors&);

  /** Hidden assignment operator. */
  Sensors& operator= (const Sensors&);

  /** The Side parsed from the current sensor data. */
  Side parsedSide;

  /** The PlayMode parsed from the current sensor data. */
  PlayMode parsedPlayMode;

  /** The ViewQuality parsed from the current sensor data. */
  ViewQuality parsedViewQuality;

  /** The ViewWidth parsed from the current sensor data. */
  ViewWidth parsedViewWidth;

  /** The Locations parsed from the current sensor data. */
  std::vector<FieldLocation> parsedLocations;

  /** The FlagOffset parsed from the current sensor data. */
  UIntUnknown parsedFlagOffset;

  /** The uniform number parsed from the current sensor data. */
  UIntUnknown parsedUniformNumber; 

  /** If the player is a goalie according to the current sensor data. */
  bool parsedGoalie;

  /** The team parsed from the current sensor data. */
  Team parsedTeam;

  /** The markers parsed in the current message. */
  std::vector<ServerMarkerObject> parsedMarkers;

  /** The lines parsed in the current message. */
  std::vector<ServerLineObject> parsedLines;

  /** The players parsed in the current message. */
  std::vector<ServerPlayerObject> parsedPlayers;
  
  /** The ball parsed in the current message. */
  ServerBallObject parsedBall;

  /** The agent parsed in the current (sense) message. */
  ServerAgentObject parsedAgent;

  /** The players observed by the coach parsed in the current message. */
  std::vector<CoachPlayerObject> parsedCoachPlayers;

  /** The ball observed by the coach parsed in the current message. */
  CoachBallObject parsedCoachBall;
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "Sensors.icc"
#endif


#endif // !SENSORS_H

