/* -*- Mode: C++ -*-
 * $Id: Dribble.cc,v 1.4 2002/10/02 14:03:15 aip7 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Dribble.cc
 * Unitname: Framework/skill
 * $Revision: 1.4 $
 * Created by: Anders och Erik
 * Last modified by $Author: aip7 $ $Date: 2002/10/02 14:03:15 $
 *
 *
 * HISTORY
 * 0.1 Initial version
 *
 *
 */

/**
 * @file Dribble.cc
 * This is a definition file for the skill Dribble.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.4 $
 * $Author: aip7 $
 * $Date: 2002/10/02 14:03:15 $
 */

#include "Dribble.h"
#include "KickCommand.h"
#include "WorldFacts.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "BallView.h"
#include "KickBallView.h"
#include "view_constants.h"


RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---

Dribble::Dribble(const Point def_dribblePoint,
		 const Float def_minDistanceTo)
  : facts(WorldFacts::instance()),
    ballView(NULL),
    agentView(NULL),
    playerView(NULL),
    dribblePoint(def_dribblePoint),
    minDistanceTo(def_minDistanceTo) {
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  getView(kickBallView, KICKBALL_VIEW_ID);
}


// --- DESTRUCTOR ---

Dribble::~Dribble() {
    viewManager->releaseView(BALL_VIEW_ID);
    viewManager->releaseView(AGENT_VIEW_ID);
    viewManager->releaseView(PLAYER_VIEW_ID);
    viewManager->releaseView(KICKBALL_VIEW_ID);
}

// --- PUBLIC METHODS ---

Plan Dribble::generatePlan() {
  static Command *cmd;
  Step step;

  //Setup the constants

  //Nuber of angles to use. Better be odd.
  const int NUM_ANGLES = 5; 
  
  //Increment between two angles.
  const int ANGLE_INCR = 10;

  //Maximum length to kick
  const float MAX_KICK_LENGTH = 15.0;

  //Importance settings. Only ratio matters.
  //Importance of going straight
  const int STRAIGHT_IMPORTANCE = 1;

  //Importance of enemy angle
  const int ANGLE_IMPORTANCE = 1;

  //Importance of enemy distance
  const int DISTANCE_IMPORTANCE = 1;

  PointUE ourPosition = agentView->getPosition();
  VectorUE vecToDribblePoint = VectorUE(ourPosition, dribblePoint);
  float angleToDribblePoint = vecToDribblePoint.getAngle().getDeg();
  
  Vector **angles = new Vector*[NUM_ANGLES];
  angles[0] = new Vector(angleToDribblePoint, NUM_ANGLES*STRAIGHT_IMPORTANCE);
  for (int k=1; k<NUM_ANGLES; k++) {
     if (k%2 == 1) {
	angles[k] = new Vector( k*ANGLE_INCR + angleToDribblePoint, 
				(NUM_ANGLES-k)*STRAIGHT_IMPORTANCE);
     }
     else {
	angles[k] = new Vector(-(k-1)*ANGLE_INCR + angleToDribblePoint, 
			       (NUM_ANGLES-k+1)*STRAIGHT_IMPORTANCE);
     }
  }

  //Avoid dribbling out of the field by setting godness of
  // angles leading outside the field to -1000
  for( int m=0; m<NUM_ANGLES; m++) {
    Point tmp = ourPosition.getPoint() + Vector(angles[m]->getAngle(), MAX_KICK_LENGTH);
    if( tmp.getY() > facts->SP_pitch_width/2.0 || 
	tmp.getY() < -(facts->SP_pitch_width)/2.0 ) {
      angles[m]->setLength(-1000);
    }
  }
  
  //Get opponenents
  std::vector<PlayerObject> opps = playerView->getOpponents();
  for (std::vector<PlayerObject>::iterator i = opps.begin(); i != opps.end(); i++) {
    RS_LOG (LA_DECISION, "An opp...");
    PointUE oppPosition = i->getPosition();
    if (oppPosition.isKnown()) {
      // Is the opponent within interception range?
      VectorUE agentToOppVector = VectorUE(ourPosition, oppPosition);
      for (int j=0; j<NUM_ANGLES; j++) {
	//Use vector length as angle goodness.
	float goodness = angles[j]->getLength();
	goodness += abs(agentToOppVector.getAngle().getDeg() - 
			angles[j]->getAngle().getDeg()) * ANGLE_IMPORTANCE;
	goodness += agentToOppVector.getLength().getValue() * DISTANCE_IMPORTANCE;
	angles[j]->setLength(goodness);
      }
    }
  }
   
  //Find the best angle
  float bestAngle = 0;
  float bestGoodness = -1;
  for(int l=0; l<NUM_ANGLES; l++) {
     RS_LOG ( LA_DECISION, "Angle "<< l << " " << angles[l]->getAngle().getDeg() << " " <<
	      angles[l]->getLength() );
    if(angles[l]->getLength() > bestGoodness) {
      bestGoodness = angles[l]->getLength();
      bestAngle = angles[l]->getAngle().getDeg();
    }
  }

  float kickLength = vecToDribblePoint.getLength().getValue();
  if ( kickLength > MAX_KICK_LENGTH ) {
    kickLength = MAX_KICK_LENGTH;
  }

  RS_LOG ( LA_DECISION, "Kicking to dir: " << bestAngle
	   << " length: " <<  kickLength );
  RS_LOG ( LA_DECISION, "Dribblepoint dir: " << vecToDribblePoint.getAngle().getDeg()
	   << "length: " << vecToDribblePoint.getLength().getValue() );

  cmd = new KickCommand(kickLength*3, bestAngle - agentView->getBodyDirection().getDeg());
  step.addCommand(cmd);

  //Free up used memory.
  delete [] angles;
  delete cmd;

  return Plan(step);

}

void Dribble::setDribblePoint(const Point point) {
  dribblePoint = point;
}

FuzzyBool Dribble::persistent() {
  return FuzzyBool(0.0);
}

FuzzyBool Dribble::applicable() {
  // If ball is within kick range return true.
  if (ballView->getPosition(NEXT).isKnown() &&
      ballView->getRelativeDistance(NEXT).isKnown() &&
      ballView->getRelativeDistance(NEXT) <= facts->SP_kickable_area &&
      agentView->getPosition().isKnown() &&
      agentView->getBodyDirection().isKnown() &&
      agentView->getSpeedVector().isKnown()) {
    if (abs(ballView->getPosition(NEXT).getDistanceTo(dribblePoint).getValue()) > minDistanceTo) {
      return FuzzyBool(1.0);
    }
    else {
      RS_LOG ( LA_DECISION, "Ball is already close to the dribble point.");
      return FuzzyBool(0.0);
    }
  }
  else {
    return FuzzyBool(0.0);
  }
}


RS_END_NAMESPACE
