/* -*- Mode: C++ -*-
 * $Id: DribbleToPosition.cc,v 1.15 2002/10/14 15:14:41 aip8 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: DribbleToPosition.cc
 * Unitname: Framework/skill
 * $Revision: 1.15 $
 * Created by: phigr835
 * Last modified by $Author: aip8 $ $Date: 2002/10/14 15:14:41 $
 *
 *
 * HISTORY
 * $Log: DribbleToPosition.cc,v $
 * Revision 1.15  2002/10/14 15:14:41  aip8
 * Improvement:
 * - Added weight for end position
 *
 * Revision 1.14  2002/10/09 16:05:35  aip8
 * Fixed bug:
 * - last sector wasn't initialised
 *
 * Revision 1.13  2002/10/09 15:14:19  aip8
 * Fixed bugs:
 * - Not all steps of a plan got executed
 * - Crashed when a step was empty
 *
 * Revision 1.12  2002/09/29 22:27:29  aip8
 * MoveWithBallToPosition2.cc/.h are more primitive but faster versions of MoveWithBallToPosition.cc/.h
 * The other files are implementations of skills.
 *
 * Revision 1.11  2002/09/18 14:40:45  aip8
 * Initial version of DribbleToPosition
 *
 *
 * Revision 0.0  2002/09/05 21:12:01  Rosario, Marinus, Philip
 * #include rs_log.h
 *
 */

/**
 * @file DribbleToPosition.cc
 * Implementation file for the skill class DribbleToPosition.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.15 $
 * $Author: aip8 $
 * $Date: 2002/10/14 15:14:41 $
 */


#include "WorldFacts.h"
#include "AgentView.h"
#include "BallView.h"
#include "GameView.h"
#include "PlayerView.h"
#include "view_constants.h"
#include "DribbleToPosition.h"
#include "MoveWithBallToPosition.h"
#include "MoveWithBallToPosition2.h"
#include "KickToControlBall.h"
#include "TurnCommand.h"
//#include "FindBall.h"

#include <vector>

RS_BEGIN_NAMESPACE


DribbleToPosition::DribbleToPosition(const Point& pos)
  : SkillWithPosition(pos),
    worldFacts(NULL),
    ballView(NULL),
    gameView(NULL),
    agentView(NULL),
    playerView(NULL),
    nextPositionToMoveTo(pos),
    nextSpeed(0.5f),
    kickToControlBallSkill(),
    moveWithBallToPositionSkill(nextPositionToMoveTo),
    planFinished(false),
    dribblingFailed(false) {
  worldFacts = WorldFacts::instance();
  getView(ballView, BALL_VIEW_ID);
  getView(gameView, GAME_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
}

DribbleToPosition::~DribbleToPosition() {
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}

Plan DribbleToPosition::generatePlan() {
  static int persistentSkill = DRIBBLE_TO_POSITION_NOSKILL;
  Plan plan;

  RS_LOG (LA_SKILL, "||DribbleToPosition::generatePlan():");

  // if one of the called skills is persistent call it again and return it's plan.
  if(persistentSkill != DRIBBLE_TO_POSITION_NOSKILL){
    switch(persistentSkill){

    case DRIBBLE_TO_POSITION_KICKTOCONTROLBALLSKILL:
      RS_LOG (LA_SKILL, "||  continue persistent KickToControlBall");

      plan = kickToControlBallSkill.generatePlan();

				// check if skill is persistent
      if(kickToControlBallSkill.persistent().isTrue())
				persistentSkill = DRIBBLE_TO_POSITION_KICKTOCONTROLBALLSKILL;
      else
				persistentSkill = DRIBBLE_TO_POSITION_NOSKILL;
      RS_LOG (LA_SKILL, "||    persistent: " << (kickToControlBallSkill.persistent().isTrue() ? "true" : "false"));
      return plan;

    case DRIBBLE_TO_POSITION_MOVEWITHBALLTOPOSITIONSKILL:
      RS_LOG (LA_SKILL, "||  continue persistent MoveWithBallToPosition");

      plan = moveWithBallToPositionSkill.generatePlan();

				// check if skill is still persistent
      if(moveWithBallToPositionSkill.persistent().isTrue())
				persistentSkill = DRIBBLE_TO_POSITION_MOVEWITHBALLTOPOSITIONSKILL;
      else
				persistentSkill = DRIBBLE_TO_POSITION_NOSKILL;
      RS_LOG (LA_SKILL, "||    persistent: " << (moveWithBallToPositionSkill.persistent().isTrue() ? "true" : "false"));

      return plan;


    default:
      RS_LOG (LA_SKILL, "||  unknown persistent skill, continuing normal ");
      persistentSkill = DRIBBLE_TO_POSITION_NOSKILL;
    }
  }

  // check whether ball is already under control
  if(moveWithBallToPositionSkill.applicable().isTrue()){

    RS_LOG (LA_SKILL, "||  MoveWithBallToPosition");

    // Ball is under control
    PointUE currentAgentPosition = agentView->getPosition();

    RS_LOG (LA_SKILL, "||    current agent position is known: " << (currentAgentPosition.isKnown() == true ? "yes" : "no"));

    // try to find next position to move to
    if(currentAgentPosition.isKnown()){

      // check opponents so the agent can avoid them.
      std::vector<PlayerObject> opponents = playerView->getOpponents(NEXT);
      if(opponents.empty() || !agentView->getBodyDirection().isKnown()){
				 RS_LOG (LA_SKILL, "||    see any opponents: no");
				// no known opponents
				// just go towards end position half as far as we can see opponents

				// get vector to end position
				CoordVector<Float> vector = currentAgentPosition.getPoint().getVectorTo(getPosition());
				// if end position is within save distance, go there
				if(vector.getLength() <= worldFacts->SP_team_far_length / 2){
				  nextPositionToMoveTo = getPosition();
				}else{
				  vector.setLength(worldFacts->SP_team_far_length / 2);
				  nextPositionToMoveTo = currentAgentPosition.getPoint() + vector;
				}

      }else{
				RS_LOG (LA_SKILL, "||    see any opponents: yes");
				Float sectors[DRIBBLE_TO_POSITION_SECTORS + 2];
				Point sectorEndPoints[DRIBBLE_TO_POSITION_SECTORS];

				// initialise sector end points
				Float length = worldFacts->SP_team_far_length / 2;
				for(int i = 0; i < DRIBBLE_TO_POSITION_SECTORS; i++){
				  Float angle = (((i + 0.5f) * 2 * DRIBBLE_TO_POSITION_MAXTURNANGLE) / DRIBBLE_TO_POSITION_SECTORS) - DRIBBLE_TO_POSITION_MAXTURNANGLE;
				  sectorEndPoints[i] = currentAgentPosition.getPoint() + CoordVector<Float>(angle, length);
				}

				// initialise sectors
				sectors[0] = 1000.0f;
				sectors[DRIBBLE_TO_POSITION_SECTORS + 1] = 1000.0f;
				Float endPositionDegrees = currentAgentPosition.getPoint().getVectorTo(getPosition()).getAngle().getDeg();
				for(int i = 1; i < DRIBBLE_TO_POSITION_SECTORS + 1; i++){
				  Point* point = &sectorEndPoints[i - 1];
				  // if sector end point is outside the pitch or
				  // angle is more then a certain amount of degrees from the end point away
				  // then give this sector a high value
				  if(point->getX() < -(worldFacts->SP_pitch_length / 2)
				     || point->getX() > (worldFacts->SP_pitch_length / 2)
				     || point->getY() < -(worldFacts->SP_pitch_width / 2)
				     || point->getY() > (worldFacts->SP_pitch_width / 2)){
				    sectors[i] = 1000.0f;
				  }else{
				  	Float angle = abs(currentAgentPosition.getPoint().getVectorTo(*point).getAngle().getDeg() - endPositionDegrees);
				  	if( angle > DRIBBLE_TO_POSITION_MAXRELBODYANGLE){
				    	// if angle is more then a certain amount of degrees from the end position away
				    	// then give this sector a semi high value
				    	sectors[i] = 500.0f;
				  	}else{
				    	sectors[i] = 0.0f;
				  	}
				  }
				}

				// weight sectors
				for(std::vector<PlayerObject>::iterator i = opponents.begin(); i != opponents.end(); i++){
				  if(i->getPosition().isKnown()){
						// check if opponent is in one of the sectors
				    Float deg = currentAgentPosition.getPoint().getDirectionTo(i->getPosition().getPoint()).getDeg() - agentView->getBodyDirection().getDeg();
				    if(deg > -DRIBBLE_TO_POSITION_MAXTURNANGLE && deg < DRIBBLE_TO_POSITION_MAXTURNANGLE){
							// weight = 1 / distance
				      Float weight = 1 / currentAgentPosition.getPoint().getDistanceTo(i->getPosition().getPoint());
							// index of sector the opponent is in
				      UInt index = (UInt)(((deg + DRIBBLE_TO_POSITION_MAXTURNANGLE) * DRIBBLE_TO_POSITION_SECTORS) / (2 * DRIBBLE_TO_POSITION_MAXTURNANGLE));
							// give dubble weigth to the sector where the opponent is in next cycle
							// and single weight to the adjacent sectors (where the opponent also could be, but probably isn't)
				      sectors[index++] += weight;
				      sectors[index++] += 2 * weight;
				      sectors[index] += weight;
				    }
				  }
				}

				// search best sector
				UInt bestSector = 1;
				for(int i = 2; i < DRIBBLE_TO_POSITION_SECTORS + 1; i++){
					// the smaller the weight the better
				  if(sectors[i] < sectors[bestSector]) bestSector = i;
					else if(sectors[i] == sectors[bestSector]){
						// if two sectors have the same weight,
						// select the one that is closer to the end position
						Float currentBestAngle = abs(currentAgentPosition.getPoint().getVectorTo(sectorEndPoints[bestSector-1]).getAngle().getDeg() - endPositionDegrees);
						Float newAngle = abs(currentAgentPosition.getPoint().getVectorTo(sectorEndPoints[i-1]).getAngle().getDeg() - endPositionDegrees);
						if(newAngle < currentBestAngle) bestSector = i;
					}
				}

				// print out debug information
				RS_LOG (LA_SKILL, "||    Sectors:");
				RS_LOG (LA_SKILL, "||      " << 0 << ": #Point(.,.), " << sectors[0]);
				for(int i = 1; i < DRIBBLE_TO_POSITION_SECTORS + 1; i++)
	  			RS_LOG (LA_SKILL, "||      " << i << ": " << sectorEndPoints[i - 1] << ", " << sectors[i]);
				RS_LOG (LA_SKILL, "||      " << (DRIBBLE_TO_POSITION_SECTORS + 1) << ": #Point(.,.), " << sectors[DRIBBLE_TO_POSITION_SECTORS + 1]);
				RS_LOG (LA_SKILL, "||    Best sector: " << bestSector);

				// set next position to move to
				nextPositionToMoveTo = sectorEndPoints[bestSector - 1];
      }


      // generate plan
      moveWithBallToPositionSkill.setPosition(nextPositionToMoveTo);
      moveWithBallToPositionSkill.setSpeed(nextSpeed);
      RS_LOG (LA_SKILL, "||    position: " << nextPositionToMoveTo);
      RS_LOG (LA_SKILL, "||    speed: " << nextSpeed);

      plan = moveWithBallToPositionSkill.generatePlan();

      // check if skill is persistent
      if(moveWithBallToPositionSkill.persistent().isTrue())
				persistentSkill = DRIBBLE_TO_POSITION_MOVEWITHBALLTOPOSITIONSKILL;
      else
				persistentSkill = DRIBBLE_TO_POSITION_NOSKILL;
      RS_LOG (LA_SKILL, "||    persistent: " << (moveWithBallToPositionSkill.persistent().isTrue() ? "true" : "false"));

      // check if postcondition will be true after this plan
      if((nextPositionToMoveTo == getPosition() || nextPositionToMoveTo.getDistanceTo(getPosition()) < 1.0f) && persistentSkill == DRIBBLE_TO_POSITION_NOSKILL) planFinished = true;

    }

  }else if(kickToControlBallSkill.applicable().isTrue()){

    RS_LOG (LA_SKILL, "||  KickToControlBall");

    // Ball needs to get under control
    if(agentView->getPosition().isKnown()){
      // set direction towards end position
      //RS_LOG (LA_SKILL, "||    final direction (position known): " << agentView->getPosition().getPoint().getDirectionTo(getPosition()).getDeg());
      kickToControlBallSkill.setFinalDir(agentView->getPosition().getPoint().getDirectionTo(getPosition()).getDeg());
      //RS_LOG (LA_SKILL, "||    after setFinalDir");
    }else{
      // position is unknown; is this a good idea?
      //RS_LOG (LA_SKILL, "||    final direction (position unknown): " << Point(0.0f, 0.0f).getDirectionTo(getPosition()).getDeg());
      kickToControlBallSkill.setFinalDir(Point(0.0f, 0.0f).getDirectionTo(getPosition()).getDeg());
      //RS_LOG (LA_SKILL, "||    after setFinalDir");
    }
    //RS_LOG (LA_SKILL, "||    before generatePlan");
    plan = kickToControlBallSkill.generatePlan();
    //RS_LOG (LA_SKILL, "||    after generatePlan");

    // check if skill is persistent and plan is not yet finished
    if(kickToControlBallSkill.persistent().isTrue()){
      persistentSkill = DRIBBLE_TO_POSITION_KICKTOCONTROLBALLSKILL;
    }else{
      persistentSkill = DRIBBLE_TO_POSITION_NOSKILL;
    }
    RS_LOG (LA_SKILL, "||    persistent: " << (kickToControlBallSkill.persistent().isTrue() ? "true" : "false"));

  }else{

    RS_LOG (LA_SKILL, "||  failed");

    // ball seems to be out of reach -> dribbling failed
    Command* cmd = new TurnCommand(0);
    Step* step = new Step(cmd);
    plan.addStep(*step);
    delete step;
    delete cmd;
    planFinished = true;
    dribblingFailed = true;
  }

  return plan;
}

FuzzyBool DribbleToPosition::persistent(){
  return FuzzyBool(planFinished ? 0.0f : 1.0f);
}

/** Return @c true (close to 1) if the skill has succeeded,
 *  i.e. its post conditions are satisfied.
 */
FuzzyBool DribbleToPosition::succeed(){
  return FuzzyBool(planFinished ? 1.0f : (dribblingFailed ? 0.0f : 0.5f));
}

/** Return @c true (close to 1) if the skill has failed,
 *  i.e. something went wrong.
 */
FuzzyBool DribbleToPosition::failed(){
  return FuzzyBool(dribblingFailed ? 1.0f : (planFinished ? 0.0f :0.5f));
}

FuzzyBool DribbleToPosition::applicable(){
  return (kickToControlBallSkill.applicable() || moveWithBallToPositionSkill.applicable())
    && FuzzyBool(agentView->getPosition().isKnown()
		 && getPosition().getDistanceTo(agentView->getPosition().getPoint()) > 3
		 ? 1.0f : 0.0f)
    ;
}


RS_END_NAMESPACE
