/* -*- Mode: C++ -*-
 * $Id: FreeKick.cc,v 1.2 2002/10/14 10:52:10 aip9 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FreeKick.cc
 * Unitname: Framework/skill
 * $Revision: 1.2 $
 * Created by: Fredrik Heintz 2000-09-07
 * Last modified by $Author: aip9 $ $Date: 2002/10/14 10:52:10 $
 *
 *
 * HISTORY
 *
 * $Log: FreeKick.cc,v $
 * Revision 1.2  2002/10/14 10:52:10  aip9
 * Update skill
 *
 * Revision 1.1  2002/09/27 15:41:02  aip9
 * Added the FreeKick and FreeKickPositioning skill
 *
 * Revision 1.5  2002/09/13 14:48:56  s02davro
 * FALSE -> FuzzyBool(0.0), TRUE -> FuzzyBool(1.0)
 *
 * Revision 1.4  2001/09/06 13:13:39  andbo928
 * Added class specifier to a lot of methods. Added succeed and failed methods to cc file.
 *
 * Revision 1.3  2000/09/16 18:39:39  frehe
 * Fixed the log messages.
 *
 * Revision 1.2  2000/09/16 18:36:48  frehe
 * Made minor fixes the skill stubs.
 *
 * Revision 1.1  2000/09/07 16:10:34  frehe
 * Added stubs for the skills.
 *
 */

/**
 * @file FreeKick.cc
 * Implementation file for the skill class FreeKick.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.2 $
 * $Author: aip9 $
 * $Date: 2002/10/14 10:52:10 $
 */


#include "FreeKick.h"
#include "view_constants.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "DashCommand.h"


#define FREEKICK_MOVE_TO_FREEKICK_POINT 0
#define FREEKICK_TURN_TO_TARGET 1
#define FREEKICK_IDLE 2


RS_BEGIN_NAMESPACE


// --- CONSTRUCTORS ---

/**
 * This class describes the FreeKick skill. When a free kick is called the closest player runs to a
 * "freekickPoint" behind the ball. (facing opponents goal as default). The agent turns his head to
 * get a good overview of the field and looks for a smart pass. If he does not find a good pass 
 * within a certain time (parameter maxTime) he makes his best effort.  
 *
 * Preconditions
 *  The skill is applicable when...
 *   - own freekick is called AND 
 *   - the position of the ball is known AND
 *   - the position of the agent is known. 
 * Postconditions
 *  The skill is "dropped" whenever...
 *   - an agent discovers that he is not closest to the point where the free kick was called.
 *   - the time passed since the free kick was called is greater than the maxTime-parameter.  
 *   - the agent finds a pass which is good enough. (not implemented yet. Will use some pass-skill)
 */  
FreeKick::FreeKick(const Float32 def_ballPlayerDistance,
			const Int32 def_maxTime,
			const Int32 def_dash,
			const Float32 def_maxDistance,
			const AngleDeg def_minTurnAngle)
	: facts(WorldFacts::instance()), 
	agentView(NULL),
	playerView(NULL),
	ballView(NULL),
	gameView(NULL),
	startTime(0),
	state(FREEKICK_MOVE_TO_FREEKICK_POINT),
	ballPlayerDistance(def_ballPlayerDistance),
	maxTime(def_maxTime),
	dash(def_dash),
	maxDistance(def_maxDistance),
	minTurnAngle(def_minTurnAngle)
{
	getView(agentView, AGENT_VIEW_ID);
	getView(playerView, PLAYER_VIEW_ID);
	getView(ballView, BALL_VIEW_ID);
	getView(gameView, GAME_VIEW_ID);

	target = facts->getTheirGoal().getPosition();
}

FreeKick::FreeKick(const Point def_target,
			const Float32 def_ballPlayerDistance,
			const Int32 def_maxTime,
			const Int32 def_dash,
			const Float32 def_maxDistance,
			const AngleDeg def_minTurnAngle)
	: facts(WorldFacts::instance()), 
	agentView(NULL),
	playerView(NULL),
	ballView(NULL),
	gameView(NULL),
	startTime(0),
	state(FREEKICK_MOVE_TO_FREEKICK_POINT),
	ballPlayerDistance(def_ballPlayerDistance),
	maxTime(def_maxTime),
	dash(def_dash),
	maxDistance(def_maxDistance),
	minTurnAngle(def_minTurnAngle),
	target(def_target)
{
	getView(agentView, AGENT_VIEW_ID);
	getView(playerView, PLAYER_VIEW_ID);
	getView(ballView, BALL_VIEW_ID);
	getView(gameView, GAME_VIEW_ID);
}


// --- DESTRUCTOR ---

FreeKick::~FreeKick()
{
	viewManager->releaseView(AGENT_VIEW_ID);
	viewManager->releaseView(PLAYER_VIEW_ID);
	viewManager->releaseView(BALL_VIEW_ID);
	viewManager->releaseView(GAME_VIEW_ID);
}


// --- METHODS ---

Plan FreeKick::generatePlan()
{
	static Command * cmd;
	Step step;

	switch (state)
	{
		case FREEKICK_MOVE_TO_FREEKICK_POINT:
		{
			// calculate freekickPoint, which is a point on the extension of an imaginary
			//line between opponents goal and ball. The "extensionfactor" is the distance
			//between the freekickPoint and the ball.
			
			Point ballPoint = ballView->getPosition().getPoint();
			Vector targetBallVector = target.getVectorTo(ballPoint);
			Float32 targetBallDistance = targetBallVector.getLength();
			Float32 extensionfactor = pow(1 + (ballPlayerDistance / targetBallDistance), 2);
			Point freekickPoint = target + (targetBallVector * extensionfactor);

			// calcualte "error", the distance between the freekickPoint and the agent.
			Float32 error = freekickPoint.getDistanceTo(agentView->getPosition().getPoint());


			// align neck with body
			cmd = new TurnNeckCommand((agentView->getBodyDirection() - 
				agentView->getNeckDirection()).getDeg());

			step.addCommand(cmd);
			delete cmd;

			// move to freekickPoint
			AngleDeg turnAngle = (agentView->getPosition().getDirectionTo(freekickPoint) -
							agentView->getBodyDirection()).getAngle();

			if (abs(turnAngle) > minTurnAngle)
				cmd = new TurnCommand(turnAngle);
			else
				if (error < 2)
					cmd = new DashCommand(20);
				else
					cmd = new DashCommand(dash);

			if (error < maxDistance)
				state = FREEKICK_TURN_TO_TARGET;

			break;
		}
		
		case FREEKICK_TURN_TO_TARGET:
		{
			// Turn body to face "target". (Opponents goal as default)
			AngleDeg freekickTurnAngle = (agentView->getPosition().getDirectionTo(target) -
								agentView->getBodyDirection()).getAngle();

			cmd = new TurnCommand(freekickTurnAngle);

			state = FREEKICK_IDLE;
			break;
		}

//		case FREEKICK_IDLE:
		default:
		{
			// turn head back and forth to get a good overview of field.
			AngleDeg turnAngle = 0;

			if (gameView->getNextSeeTime() == gameView->getNextTime())
				if (agentView->getNeckDirection().getDeg() < -80)
					turnAngle = 180;
				else
					if (agentView->getViewWidth() == VW_NARROW)
						turnAngle = -45;
					else
						turnAngle = -90;

			cmd = new TurnNeckCommand(turnAngle);
		}
	}  

	step.addCommand(cmd);
	delete cmd;

	return Plan(step);
}

FuzzyBool FreeKick::persistent()
{
	// Never persistent.
	return FuzzyBool(0.0);
}

FuzzyBool FreeKick::applicable()
{
	if (gameView->getPlayMode() == PM_OUR_FREE_KICK)
	{
		Time currentTime = gameView->getCurrentTime();

		if (startTime == 0)
			startTime = currentTime;

		if ((currentTime - startTime < maxTime) &&
			ballView->getPosition().isKnown() &&
			agentView->getPosition().isKnown())
		{
			PointUE player_point = playerView->getTeammateClosestToBall().getPosition();
			
			// check if agent is closest to the ball.
			if (player_point.isKnown())
			{
				Point ballPoint = ballView->getPosition().getPoint();
				Float32 pd = player_point.getDistanceTo(ballPoint).getValue();
				Float32 ad = ballView->getRelativeDistance().getValue();

				if (pd > ad)
				{
					return FuzzyBool(1.0);
				}
			}
			else // if player isUnknown
			{
				return FuzzyBool(1.0);
			}
		}
	}
	else
	{
		startTime = 0;
		state = FREEKICK_MOVE_TO_FREEKICK_POINT;
	}
  
	return FuzzyBool(0.0);
}

FuzzyBool FreeKick::succeed()
{
	// Always succedes.
	return FuzzyBool(0.0);
}

FuzzyBool FreeKick::failed()
{
	// Never fails.
	return FuzzyBool(0.0);
}


RS_END_NAMESPACE
