/* -*- Mode: C++ -*-
 * $Id: GotoPassablePosition.cc,v 1.15 2002/10/13 20:38:52 aip12 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: GotoPassablePosition.cc
 * Unitname: Framework/skill
 * $Revision: 1.15 $
 * Created by: Silke and Kai (aip12)
 * Last modified by $Author: aip12 $ $Date: 2002/10/13 20:38:52 $
 *
 *
 * HISTORY
 *
 */

/**
 * @file GotoPassablePosition.cc
 * Implementation file for the skill class GotoPassablePosition.
 * @pre The preconditions of the skill are...
 * - position of ball must be known
 * - the player doesn't have the ball
 * - the ball must be in possession of a teammate
 * - there must be a reachable passable position
 * - the current position must not be passable
 * - there must be a passable position in reach, 
 *   which is on the pitch and not in offside*
 * @post The postconditions for the skill are...
 * - the player is in a position where an interception of
 *   the ball by an enemy player is unlikely
 * - the calculated position is on the pitch
 * - the calculated position is not in offside
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.15 $
 * $Author: aip12 $
 * $Date: 2002/10/13 20:38:52 $
 */

#include "MoveToPosition.h"
#include "GotoPassablePosition.h"
#include "BallView.h"
#include "PlayerView.h"
#include "GameView.h"
#include "WorldFacts.h"
#include "view_constants.h"
#include "TurnCommand.h"
#include "DashCommand.h"


RS_BEGIN_NAMESPACE

// two angles that span a subspace
// By definition, the first member must always be counterclockwise to
// the second.
// AngleRadPair(0,0) is an invalid pair
// typedef std::pair<AngleRad,AngleRad> AngleRadPair;

/**
 * Constructor.
 */
GotoPassablePosition::GotoPassablePosition(const Int def_dash,
					   const AngleDeg def_minTurnAngle) /* check params */
  : facts(WorldFacts::instance()),
    ballView(NULL), 
    playerView(NULL),
    agentView(NULL),
    gameView(NULL) {
  getView(ballView, BALL_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(gameView, GAME_VIEW_ID);
  setMaxDistance(facts->SP_kickable_margin); /* check */
  setMinTurnAngle(def_minTurnAngle);         /* check */
  setDash(def_dash);                         /* check */
} 

/**
 * Destructor.
 */
GotoPassablePosition::~GotoPassablePosition() {
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
}

/**
 * Returns whether angle a2 lies clockwise to a1. 
 */
bool GotoPassablePosition::
isClockWise(AngleRad a1, AngleRad a2) {
  if (a1 >= 0)
    if (a2 >= 0)
      return (a2 > a1);
    else // a2 < 0
      return !(a2 + PI > a1);
  else // a1 < 0
    if (a2 < 0)
      return (a2 > a1);
    else // a2 >= 0
      return !(a2 - PI > a1);
}

/**
 * Calculates the angles of the tangents from the ball 'point' to the circle
 * around the agent 'circleCenter' with radius 'circleRad'.
 */
AngleRadPair GotoPassablePosition::
getTangentAnglesFromPointToCircle(Point point,
				  Point circleCenter,
				  float circleRadius) {
  float d,r1,r2,a,h,m1,m2,q1,q2,p1,p2,s1,s2,t1,t2;
  AngleRad temp;

  RS_LOG(LA_SKILL, "GotoPassablePosition circle: (" 
	 << circleCenter << "," << circleRadius << ")");
  p1 = point.getX();
  p2 = point.getY();
  m1 = circleCenter.getX();
  m2 = circleCenter.getY();

  r1 = circleRadius;
  
  d = sqrt(sqr(m1-p1)+sqr(m2-p2));
  if(d<r1){
    // point is in the circle
    return AngleRadPair(0,0);
  }

  // calculate intersection points of the circle around 'point' (ball) with 
  // the distance to 'circleCenter' (agent) as radius and the circle around 
  // 'circleCenter'  
  r2 = sqrt(sqr(d)-sqr(r1));
  a = (sqr(r2)-sqr(r1)+sqr(d))/(2*d);
  h = sqrt(sqr(r1)-sqr(d-a));
  q1 = p1+ (a*(m1-p1))/d;
  q2 = p2+ (a*(m2-p2))/d;

  s1 = q1 + (h*(m2-p2))/d;
  s2 = q2 - (h*(m1-p1))/d;

  t1 = q1 - (h*(m2-p2))/d;
  t2 = q2 + (h*(m1-p1))/d;

  // return tangents
  AngleRadPair arp = AngleRadPair(point.getDirectionTo(Point(s1,s2)).getRad(),
				  point.getDirectionTo(Point(t1,t2)).getRad());

  // make sure the second tangent is clockwise to the first
  if (!isClockWise(arp.first,arp.second)) {
    temp = arp.first;
    arp.first = arp.second;
    arp.second = temp;
  }

  return arp;
}

/**
 * Reduces the AngleRadPair 'a1' by the AngleRadPair 'a2'.
 */
AngleRadPair GotoPassablePosition::
subtractRadPairs(AngleRadPair a1, AngleRadPair a2) {
  // Complete occlusion, we cannot go anywhere
  if ((!isClockWise(a1.first,a2.first)) && (isClockWise(a1.second,a2.second)))
    return AngleRadPair(0,0);
  
  // Partial occlusion
  if ((isClockWise(a1.first,a2.first)) && (!isClockWise(a1.second,a2.first)))
    a1.second = a2.first;
  if ((isClockWise(a1.first,a2.second)) && (!isClockWise(a1.second,a2.second)))
    a1.first = a2.second;
  return a1;
}

/**
 * Calculates an AngleRadPair which defines a corridor which consists
 * passable positions.
 */
AngleRadPair GotoPassablePosition::
calcPassableCorridor(Point agentPosition,
		     Point ballPosition,
		     std::vector<Point> opponentPoints) {
  AngleRadPair agentRadPair;
  float opponentDistance;

  float agentSpeed = facts->SP_player_speed_max;
  float agentDistance = agentPosition.getDistanceTo(ballPosition);

  float ballSpeed = agentDistance > 50 ? facts->SP_ball_speed_max : 
    (0.015 * agentDistance + 0.25) * facts->SP_ball_speed_max;

  agentRadPair = getTangentAnglesFromPointToCircle(ballPosition,
					      agentPosition,
					      agentDistance * 
						   agentSpeed / ballSpeed); 
  // if the corridor of an opponent intersects the agent's corridor we must
  // narrow the agent's corridor
  for (std::vector<Point>::iterator i = opponentPoints.begin();
       i != opponentPoints.end(); i++) {
    opponentDistance = i->getDistanceTo(ballPosition);
    if (opponentDistance < agentDistance) {
      agentRadPair = 
	subtractRadPairs(agentRadPair,
			 getTangentAnglesFromPointToCircle(ballPosition,
							   *i,
							   opponentDistance * agentSpeed / ballSpeed));
    }
  }
  // the angles in agentRadPair specifies the corridor consisting all 
  // positions which cannot intercept by an opponent
  return agentRadPair;
}

/**
 * Calculates the boundaries of the field reduced by the team's offside line
 */   
void GotoPassablePosition::
calcBoundaries(Float *bounds) {

  std::vector<PlayerObject> opponentObjects = playerView->getOpponents(NEXT);
  
  bounds[0] = -facts->SP_pitch_length/2;
  bounds[1] = facts->SP_pitch_length/2;
  bounds[2] = -facts->SP_pitch_width/2;
  bounds[3] = facts->SP_pitch_width/2;

  // calculate the offside line by sorting the positions of the opponents
  if (opponentObjects.size() >= 2) {
    std::vector<Float> positions;
    for (std::vector<PlayerObject>::iterator i = opponentObjects.begin();
	 i != opponentObjects.end(); i++) 
      positions.insert(positions.end(),i->getPosition().getPoint().getX());
    
    std::sort(positions.begin(),positions.end());

    if (gameView->getTheirSide() == S_RIGHT) 
      bounds[1] = positions[positions.size() - 2];
    else
      bounds[0] = positions[1];
  }
}

/**
 * Returns 'true' if there exists a position where the ball can be passed without being intercepted 
 * by an opponent.
 */ 
bool GotoPassablePosition::
calcPassablePosition(Point agentPosition,
		     Point ballPosition,
		     std::vector<Point> opponentPoints) {
  // calculate the corridor insisting all positions which cannot intercept by an opponent
  AngleRadPair arp = calcPassableCorridor(agentPosition, ballPosition, 
					  opponentPoints);
  // if the corridor is empty, there is no passable position
  if ((arp.first == 0) && (arp.second == 0))
      return false;
  else {
    // check if the agent is already on a passable position
    AngleRad directionToBall = ballPosition.getDirectionTo(agentPosition);
    if ((!isClockWise(directionToBall, arp.first)) && 
	(isClockWise(directionToBall, arp.second))) 
      return false;
    
    else {
      // define the closest point to the centre line of the corridor  
      AngleRad center = (arp.first + arp.second)/2;
      AngleRad normal = (center + PI/2);
      float n1 = cos(normal);
      float n2 = sin(normal);
      float c1 = cos(center);
      float c2 = sin(center);
      float s = (ballPosition.getY()-agentPosition.getY()
		 -(ballPosition.getX()-agentPosition.getX())*n2/n1)/(-c2+c1*n2/n1);
      
      Point p = Point(ballPosition.getX()+s*c1,ballPosition.getY()+s*c2);
      
      // ensure the point is on the field which is reduced by the team's 
      // offside line
      Float bounds[4];
      calcBoundaries(bounds);
      if ((p.getX() < bounds[0]) || 
	  (p.getX() > bounds[1]) || 
	  (p.getY() < bounds[2]) || 
	  (p.getY() > bounds[3])) {
	RS_LOG( LA_SKILL, "GotoPassablePosition p: " << p.getX() << "/" << p.getY());
	RS_LOG( LA_SKILL, "GotoPassablePosition Punkt verworfen! ");
	return false;
      }
      setPoint(p);
      RS_LOG( LA_SKILL, "GotoPassablePosition p: " << p.getX() << "/" << p.getY());
      
      return true;
    }
  }
}
    
/**
 * Generates a plan.
 */
Plan GotoPassablePosition::generatePlan() {
  // Move to the calculated point
  return MoveToPosition::generatePlan();
}
/**
 * Checks if this skill should be called again (this cycle).
 */
FuzzyBool GotoPassablePosition::persistent() {
  return FuzzyBool(0.0);
}

/**
 * Checks if the preconditions are true.
 */
FuzzyBool GotoPassablePosition::applicable() {
  PlayerObject teammateClosestToBall;
  PlayerObject opponentClosestToBall;
  std::vector<Point> opponentPoints;
  std::vector<PlayerObject> opponentObjects;
  
  
  // position of ball must be known
  if (ballView->getPosition().isUnknown()) 
    return false;
  //   RS_LOG(LA_SKILL, "position of ball is known");
  
  // our agent must not be in possession of the ball
  // the distance of the closest teammate to the ball must therefore
  // no greater than our agents distance to the ball
  teammateClosestToBall = playerView->getTeammateClosestToBall(NOW);
  if (teammateClosestToBall.getPosition().getDistanceTo(ballView->getPosition()) >
      agentView->getPosition().getDistanceTo(ballView->getPosition()))
    return false;
  //  RS_LOG(LA_SKILL, "a teammate is closer to the ball than i am");

  
  // a teammate must be in possession of the ball
  // that means a teammate must be closest to the ball
  opponentClosestToBall = playerView->getOpponentClosestToBall(NOW);
  if (teammateClosestToBall.getPosition().getDistanceTo(ballView->getPosition()) >
      opponentClosestToBall.getPosition().getDistanceTo(ballView->getPosition()))
    return false;
  //  RS_LOG(LA_SKILL, "a teammate is in possession of the ball");
  
  
  // determine a good position to move to
  // put all the known positions of opponents into a new vector
  opponentObjects = playerView->getOpponents(NEXT);
  for (std::vector<PlayerObject>::iterator i = opponentObjects.begin(); 
       i != opponentObjects.end() ; i++)
    opponentPoints.insert(opponentPoints.end(),i->getPosition().getPoint());
  
  // return whether we were able to calculate a passable position
  return calcPassablePosition(agentView->getPosition(NEXT).getPoint(),
			      ballView->getPosition(NEXT).getPoint(),
			      opponentPoints);
}

RS_END_NAMESPACE





