/* -*- Mode: C++ -*-
 * $Id: GotoPassablePosition.h,v 1.13 2002/10/13 20:38:52 aip12 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: GotoPassablePosition.h
 * Unitname: Framework/skill
 * $Revision: 1.13 $
 * Created by: Silke and Kai (aip12)
 * Last modified by $Author: aip12 $ $Date: 2002/10/13 20:38:52 $
 *
 *
 * HISTORY
 *
 */

/**
 * @file GotoPassablePosition.h
 * Include file for the skill class GotoPassablePosition.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.13 $
 * $Author: aip12 $
 * $Date: 2002/10/13 20:38:52 $
 */


#ifndef GOTOPASSABLEPOSITION_H
#define GOTOPASSABLEPOSITION_H

#include "MoveToPosition.h"
#include "WorldFacts.h"

RS_BEGIN_NAMESPACE

typedef std::pair<AngleRad,AngleRad> AngleRadPair;

class BallView;
class AgentView;
class PlayerView;
class GameView;

/**
 * @class GotoPassablePosition GotoPassablePosition.h
 * This class encapsulates a go to a passable position skill.
 *
 * <!--
 * @pre The preconditions of the skill are...
 * - position of ball must be known
 * - the player doesn't have the ball
 * - the ball must be in possession of a teammate
 * - there must be a reachable passable position
 * - the current position must not be passable
 * - there must be a passable position in reach, 
 *   which is on the pitch and not in offside*
 * @post The postconditions for the skill are...
 * - the player is in a position where an interception of
 *   the ball by an enemy player is unlikely
 * - the calculated position is on the pitch
 * - the calculated position is not in offside
 * @todo This needs to be done...
 * -->
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.13 $
 * $Author: aip12 $
 * $Date: 2002/10/13 20:38:52 $
 */

class GotoPassablePosition : public MoveToPosition {
public:
  /**
   * Constructor.
   */
  GotoPassablePosition(const Int def_dash = 50,
		       const AngleDeg def_minTurnAngle = AngleDeg(5.0));
  
  /**
   * Destructor.
   */
  virtual ~GotoPassablePosition();
  
  /**
   * Generates a plan.
   */
  virtual Plan generatePlan();

  /**
   * Checks if this skill should be called again (this cycle).
   */
  virtual FuzzyBool persistent();

  /**
   * Checks if the preconditions are true.
   */
  virtual FuzzyBool applicable();

protected:  
  const WorldFacts* facts;
  BallView* ballView;
  PlayerView* playerView;
  AgentView* agentView;
  GameView* gameView;

  /**
   * Calculates the angles of the tangents from the ball 'point' to the circle
   * around the agent 'circleCenter' with radius 'circleRad'.
   */
  AngleRadPair getTangentAnglesFromPointToCircle(Point point,
						 Point circleCenter,
						 float circleRadius);

  /**
   * Returns whether angle a2 lies clockwise to a1. 
   */
  bool isClockWise(AngleRad a1, AngleRad a2);

  /**
   * Reduces the AngleRadPair 'a1' by the AngleRadPair 'a2'.
   */
  AngleRadPair subtractRadPairs(AngleRadPair a1, AngleRadPair a2);

  /**
   * Calculates an AngleRadPair which defines a corridor which consists
   * passable positions.
   */
  AngleRadPair calcPassableCorridor(Point AgentPosition,
				    Point ballPosition,
				    std::vector<Point> opponentPoints);

  /**
   * Returns 'true' if there exists a position where the ball can be passed without being intercepted 
   * by an opponent.
   */ 
  bool calcPassablePosition(Point agentPosition,
			    Point ballPosition,
			    std::vector<Point> opponentPoints);

  /**
   * Calculates the boundaries of the field reduced by the team's offside line
   */   
  void calcBoundaries(Float *bounds);


};


RS_END_NAMESPACE


#endif // !GOTOPASSABLEPOSITION_H
