
/**
 * @file GotoBall.cc
 * Implementation file for the skill class GotoBall.
 *
 * @par Unit:
 *      Framework/skill
 */
#include "Mark.h"


using namespace std;

RS_BEGIN_NAMESPACE

Mark::Mark(const Rectangle& rect,
	   const float _dist_to_marked,
	   const bool _mark_outside_zone,
	   const MarkLocation _mark_where,
	   const Int max_speed)
   : SkillWithZone(rect),
     agentView(NULL),
     playerView(NULL),
     ballView(NULL),
     gameView(NULL),
     world_facts(WorldFacts::instance())
{
   getView(playerView, PLAYER_VIEW_ID);
   getView(agentView, AGENT_VIEW_ID);
   getView(ballView, BALL_VIEW_ID);
   getView(gameView, GAME_VIEW_ID);
   move_skill.setDash(max_speed);

   mark_where = _mark_where;
   dist_to_marked = _dist_to_marked;
   is_marking = false;
   mark_outside_zone = _mark_outside_zone;
}

Mark::~Mark()
{
   viewManager->releaseView(PLAYER_VIEW_ID);
   viewManager->releaseView(AGENT_VIEW_ID);
   viewManager->releaseView(BALL_VIEW_ID);
}

// Selectors

float Mark::getDistanceToMarked()
{
  return dist_to_marked;
}

bool Mark::getMarkOutsideZone()
{
  return mark_outside_zone;
}

MarkLocation Mark::getMarkLocation()
{
  return mark_where;
}

PlayerObject Mark::getMarkedOpponent()
{
  return markedOpponent;
}

bool Mark::isMarking()
{
  return is_marking;
}

//Modifiers

void Mark::setDistanceToMarked(float distance)
{
  dist_to_marked = distance;
}

void Mark::setMarkOutsideZone(bool mark)
{
  mark_outside_zone = mark;
}

void Mark::setMarkLocation(MarkLocation where)
{
  mark_where = where;
}

void Mark::setMarkedOpponent(PlayerObject opp)
{
  markedOpponent = opp;
}

Plan Mark::generatePlan()
{
   Step step;
   
   PlayerObject opp;

   opp = playerView->getOpponentClosestToAgent();

   if (opp.getPosition().isKnown() && opp.getTeam() == T_THEIR_TEAM)
   {
     PointUE mark_point;
      //Decide if we should mark between player and goal or between 
      if (ballView->getPosition().isKnown() 
	  && mark_where == BETWEEN_PLAYER_AND_BALL)
      {
	 VectorUE line_to_ball = 
	    opp.getPosition().getVectorTo(ballView->getPosition());
	 line_to_ball.setLength(dist_to_marked);
	 mark_point = opp.getPosition() + line_to_ball;
      } 
      else	 
      {     
	 //Fallback to BETWEEN_PLAYER_AND_BALL if ball position is Unknown.
	 VectorUE line_to_goal = 
	    opp.getPosition().getVectorTo(
	       (world_facts->getOurGoal().getPosition()));
	 line_to_goal.setLength(dist_to_marked);
	 mark_point = opp.getPosition() + line_to_goal;
      }

      is_marking = true;

      move_skill.setPoint(mark_point.getPoint());
      
      return move_skill.generatePlan();
   }

   // Dummy plan to prevent compilation warnings
   return Plan(new TurnNeckCommand(0.0));
}


PlayerObject Mark::getPlayerClosestToBall() {
   PlayerObject teammate = playerView->getTeammateClosestToBall();
   PlayerObject opponent = playerView->getOpponentClosestToBall();
   BallObject ball = ballView->getBallObject();

   if (ball.getPosition().isKnown())
   {
      RS_LOG (LA_SKILL, "Ball known");
      if(teammate.getPosition().isKnown() && opponent.getPosition().isKnown())
      {
	 RS_LOG (LA_SKILL, "Opponent and Teammate known");
	 if(teammate.getPosition().getDistanceTo(ball.getPosition())
	    >= opponent.getPosition().getDistanceTo(ball.getPosition()))
	 {
	    RS_LOG (LA_SKILL, "Opponent is closest to ball");
	    return opponent;
	 }
	 RS_LOG (LA_SKILL, "Teammate is closest to ball");
	 return teammate;
      }
      else if (teammate.getPosition().isKnown())
      {
	 RS_LOG (LA_SKILL, "Teammate known");
	 return teammate;
      }
      else if (opponent.getPosition().isKnown())
      {
	 RS_LOG (LA_SKILL, "Opponent known");
	 return opponent;
      }
   }
   
   return playerView->getOpponentClosestToAgent();      
}

FuzzyBool Mark::persistent()
{
   if (is_marking)
      return FuzzyBool(0.8);
   else
      return FuzzyBool(0.0);
}

FuzzyBool Mark::applicable()
{
   if(!gameView->isDashAllowed()) return FuzzyBool(0.0);

   // Ok first let's se if we are in fact the closest team member.
   std::vector<PlayerObject> opponents =
      playerView->getOpponents();
   std::vector<PlayerObject> teammates = playerView->getTeammates();
   PlayerObject closestOpponent = playerView->getOpponentClosestToAgent();

   if(opponents.size() == 0) {
      RS_LOG (LA_SKILL, "No players");    
      return FuzzyBool(0.0);
   }

   if(closestOpponent.getPosition().isKnown())
   {
      for (std::vector<PlayerObject>::iterator iter = teammates.begin();
	   iter < teammates.end();
	   iter++)       
      {
	 if(iter->getPosition().isKnown() 
	    &&
	    iter->getPosition().getDistanceTo(closestOpponent.getPosition())
	    < 
	    agentView->getPosition().getDistanceTo(closestOpponent.getPosition()))
	 {
	    //Experimental: If we are too far away let's stop marking him.
	    is_marking = false;

	    // If we find that any teammate are closer to the opponent than
	    // we are, then we shold not mark him
	    RS_LOG(LA_SKILL, "We are not closest. Don't Mark");
	    return FuzzyBool(0.0);
	 }
      }
  
      // If opponent is outside of our zone, and we aren't commited to mark him,
      // lets not mark him.
      if(!zone.inArea(closestOpponent.getPosition().getPoint()))
      {
	 RS_LOG(LA_SKILL, "Our closest opponent is not in our zone. Don't Mark");
	 if (mark_outside_zone)
	    return FuzzyBool(1.0);
	 else
	 {
	    is_marking=false;
	    return FuzzyBool(0.0);
	 }
      }
   }

   PlayerObject closestToBall = getPlayerClosestToBall();

   /* If we think that the ball is closest to a teammate we don't really 
      know if we should
     mark. Let's be a bit uncertain */
   if (closestToBall.getTeam() == T_OUR_TEAM)
   {
      RS_LOG (LA_SKILL, "Same Team");
      return FuzzyBool(0.5);
   }
   else if (closestToBall.getTeam() == T_THEIR_TEAM)
   {
      RS_LOG (LA_SKILL, "Opponent");    
      return FuzzyBool(1.0);
   }
   else
   {
      // This should not happen?
      RS_LOG (LA_SKILL, "Team not known");
      is_marking=false;
      return FuzzyBool(0.0);
   }
}

FuzzyBool Mark::succeed()
{
   return FuzzyBool(1.0);
}

FuzzyBool Mark::failed()
{
   return FuzzyBool(0.0);
}

RS_END_NAMESPACE














