/**
 * @file Mark.h
 * Include file for the class Mark.
 *
 * @par Unit:
 *      Framework/skill
 * $Author: aip4 $
 *
 */

#ifndef MARK_H
#define MARK_H

#include "SkillWithZone.h"
#include "AgentView.h"
#include "PlayerView.h"
#include "BallView.h"
#include "WorldModelInterface.h"
#include "GameView.h"
#include "MoveToPosition.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "view_constants.h"
#include "DashCommand.h"
#include "WorldFacts.h"
#include <string>

RS_BEGIN_NAMESPACE

class AgentView;


/**
 * @file Mark.h
 * Include file fpr the skill Mark. 
 *
 * @par Unit:
 *      Framework/skill
 */

enum MarkLocation { BETWEEN_PLAYER_AND_GOAL = 1, BETWEEN_PLAYER_AND_BALL };

/**
 * @class Mark Mark.h
 *  Skill to position the agent in a good position for intercept.
 *  Is a SkillWithZone which means that it uses a zone to
 *  decide in which area the agent should mark (SkillWithZone for more
 *  complete documentation). 
 *
 * <!--
 * @pre The preconditions are: 1) I see an unmarked player
 *                             2) My team doesn't control ball
 *                             3) I am the closest teammember to the ball
 *                             4) The player is in our zone (or we are
 * following him outside of it.
 *                             5) We are not in a good mark-position already
 * @post Hopefully in a good intercept-position
 * @todo The two first preconditions should be accomplished by communication
 * between teammembers.
 * -->
 */
class Mark : public SkillWithZone
{
public:
  /** Constructor. Takes the following arguments:
    - A zone to keep inside of
    - A distance to keep from our marked opponent when marking
    - A boolean telling us if we should mark outside the given zone if needed
    - A MarkLocation defining what line we are trying to cut off (to ball, to goal)
    - A max speed (0-100) to use (maximally) when running to mark
  */
  Mark(const Rectangle& rect,
       const float _dist_to_marked = Float(3.0), 
       const bool _mark_outside_zone = true,
       const MarkLocation _mark_where = BETWEEN_PLAYER_AND_GOAL,
       const Int max_speed = 100);

  virtual ~Mark();
  
  /** Generates a plan to position ourselves in a good position */
  virtual Plan generatePlan();
  
  virtual FuzzyBool persistent();
  
  /** Determines if the skill is applicable. The preconditions are:
   *    - I'm closest to the opponent of all my teammates
   *    - Our team doesn't control the ball.
   *    - The opponent closest to me is in my zone (if not MarkOutsideZone)
   */
  virtual FuzzyBool applicable();

  /** We have succeded. The postconditions are either that we have stopped
   *  marking, which can happen if we lost track of our opponent or ran out
   *  of our designated zone, or that we are at the right place.
   */
  virtual FuzzyBool succeed();

  virtual FuzzyBool failed();

  /** Gets the distance to our marked opponent */
  float getDistanceToMarked();
  /** Am I supposed to mark outside my zone? */
  bool getMarkOutsideZone();
  /** The type of marking we are up to */
  MarkLocation getMarkLocation();
  /** The opponent we are marking */
  PlayerObject getMarkedOpponent();
  /** True if we are marking someone */
  bool isMarking();

  /** Sets the distance to keep from the marked opponent */
  void setDistanceToMarked(float distance);
  /** Determine whether we should mark outside the zone */ 
  void setMarkOutsideZone(bool mark);
  /** Set the type of marking we are up to */
  void setMarkLocation(MarkLocation where);
  /** Set the opponent to be marked */
  void setMarkedOpponent(PlayerObject opp);

protected:

  /** Get a pointer to the player closest to the ball */
  PlayerObject getPlayerClosestToBall();

  /** My view of myself */
  AgentView* agentView;
  
  /** My view of the players on the field */
  PlayerView* playerView;
  /** My view of the ball */
  BallView* ballView;
  /** My view of the game */
  GameView* gameView;
  /** Uses the MoveToPosition skill to move to the correct position */
  MoveToPosition move_skill;

  /** My facts of the world */
  const WorldFacts* world_facts;

  /** How far from the player should we mark? */
  float dist_to_marked;
  
  /** Should we continue outside of our zone? */
  bool mark_outside_zone;
  
  /** Mark beteen player our ball? */
  MarkLocation mark_where;
  /** Keep track if we have started marking earlier. In that case it might
   *  be we should continue marking even if the opponent leaves our zone. */
  bool is_marking;
  PlayerObject markedOpponent;
};

RS_END_NAMESPACE

#endif


