/* -*- Mode: C++ -*-
 * $Id: ReceivePass13.h,v 1.5 2002/10/15 21:20:08 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: ReceivePass13.h
 * Unitname: Framework/skill
 * $Revision: 1.5 $
 * Created by: Laurent and Yves 2002-09-25
 * Last modified by $Author: frehe $ $Date: 2002/10/15 21:20:08 $
 *
 *
 * HISTORY
 *
 * $Log: ReceivePass13.h,v $
 * Revision 1.5  2002/10/15 21:20:08  frehe
 * Removed unnecessary code.
 *
 * Revision 1.4  2002/10/15 12:00:43  s02davro
 * Fixed conflicts in RecivePass13
 *
 * Revision 1.3  2002/10/14 22:12:33  aip13
 * Doxygene upgrade
 *
 * Revision 1.2  2002/09/29 22:06:27  aip13
 * *** empty log message ***
 *
 * Revision 1.1  2002/09/25 08:00:37  aip13
 * Added the ReceivePass13 skill.
 *
 *
 */

/**
 * @file ReceivePass13.h
 * Include file for the skill class GotoBall.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2002/10/15 21:20:08 $
 */


#ifndef RECEIVEPASS13_H
#define RECEIVEPASS13_H


#include "Skill.h"
#include "WorldFacts.h"
#include "MoveToPosAsap.h"
#include "GotoBall.h"
#include "MoveToPosition.h"


RS_BEGIN_NAMESPACE

class AgentView;
class BallView;
class PlayerView;


/**
 * @class ReceivePass13 ReceivePass13.h
 * This class implements a skill 'receiveing a pass' in the Robosoc framework.
 * The implementation takes into account several cases that can happen in the game when receiving a pass:
 * - The ball is stopped.
 * - The ball is moving:
 *   - We test if the ball is going towards the opponent goal.
 *       (Attack 'technique')  
 *     - If yes, we calculate the position where the ball will stop naturally after being passed
 *       - If we can reach this point before an opponent, we move the agent to this point.
 *       - If we cannot reach the point, we calculate a limit point where we can just reach the ball before the closest opponent,
 *         and we move to it.
 *       ('Normal' receiving pass situation)
 *     - If the ball isn't coming towards the opponent's goal: 
 *       - We calculate a perpendicular intersection between the ball and the agent.
 *         - If the agent can be at this position before or at the same time than the ball, he moves to this point.
 *         - If the agent cannot be at the inter point on time, we try to catch the ball asap, using a prediction of its trajectory.
 * 
 * @pre The preconditions for the skill ReceivingPass13 are: 
 *          - Knowing informations about:
 *            - The ball (speed, position, relative distance to it from the agent).
 *            - The agent itself, its position.
 *            - The other players, closest teammate position and closest opponent position.
 *          - We need to know if a player from the same team is closer to the ball than us, 
 *            if this is the case the skill isn't applicable.
 *          - As we use the skill "MoveToPosAsap" many times in "ReceivingPass13", we have to test if it is applicable.
 * 
 * @post The intended postcondition for the skill ReceivingPass13 is that the agent must have "touched"/catched the ball.
 *
 * @todo To try to know better if we are in a situation to receive a pass (applicable or decision-maker).
 *       To know if a pass has been made to us (communication between agents ... ?)
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.5 $
 * $Author: frehe $
 * $Date: 2002/10/15 21:20:08 $
 */
class ReceivePass13 : public Skill {
public:
  /** Default constructor. Here the skill is initialize. */
  ReceivePass13();

  /** Destructor should clean up after the skill. */
  virtual ~ReceivePass13();


  /** This method is used to generate a plan for how the goal of
    the skill should be obtained by the agent.
    See a quite detailed content of what is done in this method in the description on the class above.
  */
  virtual Plan generatePlan();

  /** This method should return a FuzzyBool telling the caller
    if the skill wants to be called again. */
  virtual FuzzyBool persistent();

  /** This method implement the precondition(s) specified in
   the description of the skill. */
  virtual FuzzyBool applicable();

  /** Return @c true (close to 1) if the skill has succeeded,
      i.e. its post conditions are satisfied. */
  virtual FuzzyBool succeed();

  /** Return @c true (close to 1) if the skill has failed,
      i.e. something went wrong. */
  virtual FuzzyBool failed();

protected:
  
  AgentView*  agentView;
  BallView*   ballView;
  PlayerView* playerView;

  /** Pointer on MoveToPosAsap, another skill we use. */
  MoveToPosAsap* moveToPosAsap;
  
  /** To be able to access all the world constants and methods in WorldFacts. */
  const WorldFacts* facts;

private:
  /** Hide copy-constructor. */
  ReceivePass13(const ReceivePass13&);

  /** Hide assignment operator. */
  ReceivePass13& operator= (const ReceivePass13&);

  /** 
   * To tell the agent to do nothing, when for example informations are missing.
   * We use a "TurnCommand" with argument "0" to make the agent waiting.
   * @return a plan with the command "TurnCommand(0)"
   */
  Plan agentIsWaiting();

  /** 
   * Compute the intersection point between the agent and the ball in movement.
   * To do this we use a "while" loop which test an inquality between two variables (tBall and tAgent) which stand for time lengths
   * In the loop we increment one of this variable (tBall).
   * After each incrementation, we compute the distance that the ball has moved during "tBall" cycles, 
   * then we compute the distance from the agent to the "after tBall cycles" position of the ball,
   * we compute afterwards the time who would take the agent to move of this distance.
   * Finally, by comparring the two time lengths in the loop condition, 
   * we will stop when we'll have the time (and so the position) of the ball when it can be reached "exactly" by the agent.
   * 
   * @return the predicted "meeting point" of the ball and the agent.
   */
  Point agentToBallInter();
};


RS_END_NAMESPACE

#endif // !RECEIVEPASS13_H
