/* -*- Mode: C++ -*-
 * $Id: Spread.cc,v 1.4 2002/10/14 17:22:35 aip1 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Spread.cc
 * Unitname: Framework/skill
 * $Revision: 1.4 $
 * Created by: Mikael Klasson, Krille, Kim Ahlin 2002/10/04
 * Last modified by $Author: aip1 $ $Date: 2002/10/14 17:22:35 $
 *
 *
 * HISTORY
 *
 * $Log: Spread.cc,v $
 * Revision 1.4  2002/10/14 17:22:35  aip1
 * Added doxygen comments.
 *
 * Revision 1.3  2002/10/13 11:50:01  aip1
 * Cleaned up the Spread skill.
 *
 * Revision 1.2  2002/09/29 20:56:50  aip1
 * Updated the Spread skill.
 *
 * Revision 1.1  2002/09/20 14:38:57  aip1
 * Added the Spread skill.
 *
 * Revision 1.5  2002/09/13 14:48:56  s02davro
 * FALSE -> FuzzyBool(0.0), TRUE -> FuzzyBool(1.0)
 *
 * Revision 1.4  2001/09/06 13:13:39  andbo928
 * Added class specifier to a lot of methods. Added succeed and failed methods to cc file.
 *
 * Revision 1.3  2000/09/16 18:39:39  frehe
 * Fixed the log messages.
 *
 * Revision 1.2  2000/09/16 18:36:48  frehe
 * Made minor fixes the skill stubs.
 *
 * Revision 1.1  2000/09/07 16:10:34  frehe
 * Added stubs for the skills.
 *
 */

/**
 * @file Spread.cc
 * Implementation file for the skill class Spread.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.4 $
 * $Author: aip1 $
 * $Date: 2002/10/14 17:22:35 $
 */


#include "Spread.h"
#include "AgentView.h"
#include "BallView.h"
#include "GameView.h"
#include "PlayerView.h"
#include "view_constants.h"
#include "MoveCommand.h"

#define DEF_RADIUS 5.0f
#define DEF_ALLOWED_PLAYERS 1

RS_BEGIN_NAMESPACE

typedef std::vector<PlayerObject> Container;


// --- CONSTRUCTORS ---

Spread::Spread()
  // This is an initializer list
  // which means that all the objects inside the
  // class can be initialized after the colon.
  // You initialize the variable myVar with
  // myVar(constructor arguments).
  // You should also initialize the base classes, but
  // only the classes that you directly inherit from.
  // If you do not initialize a variable its default
  // constructor will be called.
  : Skill(),
    ballView(NULL),
    gameView(NULL),
    agentView(NULL),
    playerView(NULL)
{
  getView(ballView, BALL_VIEW_ID);
  getView(gameView, GAME_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID);
  getView(playerView, PLAYER_VIEW_ID);
}


// --- DESTRUCTOR ---

Spread::~Spread()
{
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
}


// --- METHODS ---

Plan Spread::generatePlan()
{
  Command * cmd;
  Step step;  
  std::vector<PlayerObject> mates;
  Container::iterator i;
  VectorUE v(0,0);
  Point agentpos;

  mates = playerView->getTeammates( NOW );
  for( i = mates.begin(); i != mates.end(); i++ ) {
    if( !i->getPosition().isKnown() ) {
      continue;
    }
    if( i->getPosition().getDistanceTo( agentView->getPosition( NOW ) ) 
	< radius ) {
      v += agentView->getPosition( NOW ).getVectorTo( i->getPosition() );
    }
  }
  v = -v;
  
  if( agentView->getPosition( NOW ).isKnown() ) {
    agentpos = agentView->getPosition( NOW ).getPoint();
  }  
  
  v.setLength( 1 );
  
  Point pt( agentpos, v.getVector() );
  cmd = new MoveCommand( pt );
  step.addCommand( cmd );
  delete cmd;
  return Plan(step);
}


FuzzyBool Spread::persistent()
{
  // Never persistent.
  return FuzzyBool(0.0);
}


FuzzyBool Spread::applicable()
{
  return applicable( DEF_RADIUS, DEF_ALLOWED_PLAYERS );
}


FuzzyBool Spread::applicable(float radius, int allowedPlayers)
{
  PlayerObject po;
  std::vector<PlayerObject> mates;
  Container::iterator i;
  UInt uclose = 0;
  
  this->radius = radius;
  
  if( gameView->getPlayMode() == PM_BEFORE_KICK_OFF ) {
    return FuzzyBool(0.0);
  }
  
  if( !agentView->getPosition( NOW ).isKnown() ) {
    return FuzzyBool( 0.0 );
  }
  
  po = playerView->getTeammateClosestToBall( NOW );
  
  if( !po.getPosition().isKnown() ) {
    return FuzzyBool( 0.0 );
  }
  
  if( !ballView->getPosition( NOW ).isKnown() ) {
    return FuzzyBool( 0.0 );
  }
  
  if( po.getPosition().getDistanceTo( ballView->getPosition( NOW ) ) >
      ballView->getRelativeDistance( NOW ) ) {
    return FuzzyBool( 0.0 );
  }
  
  
  mates = playerView->getTeammates( NOW );
  for( i = mates.begin(); i != mates.end(); i++ ) {
    if( !i->getPosition().isKnown() ) {
      continue;
    }
    if( i->getPosition().getDistanceTo( agentView->getPosition( NOW ) ) 
	< radius ) {
      if( ++uclose > allowedPlayers ) {
	return FuzzyBool( 1.0 );
      }
    }
  }
  
  return FuzzyBool( 0.0 );
}

FuzzyBool Spread::succeed()
{
  
  // Always succedes.
  return FuzzyBool(0.0);
}

FuzzyBool Spread::failed()
{
  // Never fails.
  return FuzzyBool(0.0);
}



RS_END_NAMESPACE


// Remove from here
/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Spread.icc"
#endif
// To here if you do not have any inline definitions!
