/* -*- Mode: C++ -*-
 * $Id: BasicCommandStrategy.cc,v 1.5 2002/10/09 13:29:46 aip19 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: BasicCommandStrategy.cc
 * Unitname: BasicSystem
 * $Revision: 1.5 $
 * Created by: David Rosn 2002-07-01
 * Last modified by $Author: aip19 $ $Date: 2002/10/09 13:29:46 $
 *
 *
 * HISTORY
 *
 * $Log: BasicCommandStrategy.cc,v $
 * Revision 1.5  2002/10/09 13:29:46  aip19
 * Bugfix in TurnNeckCommand
 *
 * Revision 1.4  2002/10/08 16:59:25  aip6
 * Updated the CommandStrategy to handle turnneck commands out of bounds
 *
 * Revision 1.3  2002/09/04 13:50:04  s02davro
 * *** empty log message ***
 *
 * Revision 1.2  2002/08/31 15:35:24  s02davro
 * Updated the strategies and the interface
 *
 * Revision 1.1  2002/08/02 22:00:51  s02davro
 * Created the Basic classes, needs improvements.
 *
 */

/**
 * @file BasicCommandStrategy.cc
 * Implementation file for the class BasicCommandStrategy.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.5 $
 * $Author: aip19 $
 * $Date: 2002/10/09 13:29:46 $
 */

#include "BasicCommandStrategy.h"
#include "WorldModelInterface.h"
#include "CatchCommand.h"
#include "DashCommand.h"
#include "KickCommand.h"
#include "MoveCommand.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "math_utils.h"
#include "rs_error.h"

RS_BEGIN_NAMESPACE

BasicCommandStrategy::BasicCommandStrategy() {
}

BasicCommandStrategy::~BasicCommandStrategy() { 
}

void BasicCommandStrategy::updateAfterCommand() {
  RS_LOG ( LA_STRATEGY, "BasicCommandStrategy - start");
  world = &wmi->getWorld();
  AgentObject agent = world->getAgent();
  BallObject ball = world->getBall();

  ServerCommand* last_cmd;
  if (wmi->getLastCommand() == NULL || 
      (!wmi->getLastCommand()->isServerCommand())) {
    RS_LOG ( LA_STRATEGY, "BasicCommandStrategy - nocommand - end");  
    return;
  }
  last_cmd = wmi->getLastCommand()->getServerCommand();

  CommandType lastCommandType = last_cmd->getType();
  const std::string lastServerString = last_cmd->serverString();
  
  try {
    if ( lastCommandType == CMD_Catch ) {
      // ?
    } else if ( lastCommandType == CMD_Dash ) {
      RS_LOG ( LA_STRATEGY, "BasicCommandStrategy - dash");
      const DashCommand* const cmd = last_cmd->getConstDashCommand();
      if ( world->getLastPlayMode() == PM_BEFORE_KICK_OFF ) {
	// Dash shouldn't work since you're not allowed to dash in this play 
	// mode! And some others !Todo!
      } else {
	// First we need to normalize the power, since we can't be sure 
	// that the power is actually within its bounds.
	Float power = cmd->getPower();
	Float real_power = minMax (facts->SP_minpower, 
				   power, 
				   facts->SP_maxpower);
	Float needed_power = real_power;
	// Calculate the maximum needed power the agent can use
	Float max_needed_power = -facts->SP_stamina_min;
	
	max_needed_power += agent.getStamina().getValue();
	
	// If we do a negative dash the cost is doubled.
	if ( power < 0 ) {
	  needed_power *= -2;
	}
	
	// Check if the agent has enough stamina to do the dash,
	// otherwise decrease the power so that the stamina
	// will be SP_stamina_min afterwards.
	if ( needed_power > max_needed_power ) {
	  needed_power = max_needed_power;
	  // Since the cost of negative dash is doubled half the
	  // real power if negative dash.
	  real_power = ( power < 0 ) ? needed_power/-2 : needed_power;
	}
	
 	AngleDegUE bodyDir = agent.getBodyDirection();
	
	FloatUE effort = agent.getEffort();
	
  	FloatUE acc_power = real_power * effort * facts->SP_dash_power_rate;

	// Now update the acceleration of the agent.	
  	agent.setAcceleration(VectorUE(PointUE(acc_power * cos(bodyDir),
					       acc_power * sin(bodyDir))));
	
	// Then store the power used.
  	agent.setPowerUsed(needed_power);
      }
    } else if ( lastCommandType == CMD_Kick ) {
      RS_LOG ( LA_STRATEGY, "BasicCommandStrategy - kick");
      const KickCommand* const cmd = last_cmd->getConstKickCommand();
      Float power = cmd->getPower();
      AngleDeg angle = cmd->getAngle();
      PointUE agentPos = agent.getPosition();
      PointUE ballPos = ball.getPosition();
      if ( world->getLastPlayMode() == PM_BEFORE_KICK_OFF ||
	   world->getLastPlayMode() == PM_OUR_GOAL ||
	   world->getLastPlayMode() == PM_THEIR_GOAL ||
	   world->getLastPlayMode() == PM_THEIR_OFFSIDE_KICK ) {	
	// Kick shouldn't work since you're not allowed to kick in 
	// these play modes! And others? !Tocheck!
      } else if ( ballPos.isUnknown() || agentPos.isUnknown() ||
		  ballPos.getDistanceTo(agentPos) > 
		  facts->SP_kickable_area ) {
	// Kick shouldn't work since the ball is too far away!
	// If the distance is unknown this test will also fail.
      } else {
	Float dir_diff = 0;
	dir_diff = abs(agentPos.getDirectionTo(ballPos).getDeg());
	
	Float dist_diff = agentPos.getDistanceTo(ballPos).getValue() -
	  facts->SP_player_size - facts->SP_ball_size;
	Float kick_power = facts->SP_kick_power_rate * power * 
	  (1.0 - 0.25*dir_diff/180.0 - 
	   0.25*dist_diff/facts->SP_kickable_margin);
	
	AngleDegUE ball_dir = angle + agent.getBodyDirection();
	
	ball.setAcceleration(VectorUE(PointUE(kick_power * cos(ball_dir),
					      kick_power * sin(ball_dir))));
      }
    } else if ( lastCommandType == CMD_Move ) {
      const MoveCommand* const cmd = last_cmd->getConstMoveCommand();
      // If it is successful then the player should be at the intended 
      // position. If the position is legal that is. Otherwise the position 
      // will be totally random. The speed of the agent should be zero, 
      // and the neck and body dir unchanged.
      Point pos = cmd->getDestination();
      // This is not true since it is the play mode at the execution 
      // of the command that is intresting!!!
      if (((world->getLastPlayMode() == PM_BEFORE_KICK_OFF ||
	    world->getLastPlayMode() == PM_OUR_GOAL ||
	    world->getLastPlayMode() == PM_THEIR_GOAL ) &&
	   // All players can move when before kick off and after a goal
	   // If they move within their own half.
	   (pos.getX() <= 0 && abs(pos.getY()) > facts->SP_pitch_width/2)) ||
	  (world->getLastPlayMode() == PM_OUR_GOALIE_GOT_BALL &&
	   agent.isGoalie() &&
	   // The goalie is allowed to move if it has catched the ball
	   // and it moves within its penalty area.
	   pos.getX() <= (facts->SP_penalty_area_length - 
			  facts->SP_pitch_length/2) &&
	   abs(pos.getY()) > facts->SP_penalty_area_width/2)) {
	agent.setMovePosition(pos);
      }
    } else if ( lastCommandType == CMD_Turn ) {
      RS_LOG ( LA_STRATEGY, "BasicCommandStrategy - turning");
      const TurnCommand* const cmd = last_cmd->getConstTurnCommand();
      if (agent.getSpeed().isUnknown()) {
	agent.setTurn(cmd->getAngle());
      }
      else {
	AngleDegUE turnAngle(cmd->getAngle());
	turnAngle /= (facts->SP_inertia_moment*agent.getSpeed() + 1);
	agent.setTurn(turnAngle);
      }
    } else if ( lastCommandType == CMD_TurnNeck ) {
      RS_LOG ( LA_STRATEGY, "BasicCommandStrategy - turnneck");
      const TurnNeckCommand* const cmd = last_cmd->getConstTurnNeckCommand();      
      AngleDeg neckturn = cmd->getAngle();
      if (neckturn < -180) {
	neckturn = -180;
	RS_WARNING ("Tried turning neck < -180");
      }
      if (neckturn > 180) {
	neckturn = 180;
	RS_WARNING ("Tried turning neck > 180");
      }
      AngleDeg relneckang = (agent.getNeckDirection() - agent.getBodyDirection()).getDeg();
      if ((relneckang + neckturn) < -90) {
	neckturn = relneckang - 90;
	RS_WARNING ("Tried turning neck < -90 from body");
      }
      if ((relneckang + neckturn) > 90) {
	neckturn = relneckang + 90;
	RS_WARNING ("Tried turning neck > 90 from body");
      }
      agent.setTurnNeck(neckturn);      
    }    
  } catch (RSException &excpt) {
    RS_FATAL("Caught an exception myUpdateAfterCommand: " << excpt.What());
  }
  world->setBall(ball);
  world->setAgent(agent);
  wmi->setLastCommand(NULL);
  RS_LOG ( LA_STRATEGY, "BasicCommandStrategy - end");
}

RS_END_NAMESPACE

