/* -*- Mode: C++ -*-
 * $Id: WithError.h,v 1.3 2002/10/01 16:08:39 s02davro Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: WithError.h
 * Unitname: Library
 * $Revision: 1.3 $
 * Created by: David Rosn 2002-06-19
 * Last modified by $Author: s02davro $ $Date: 2002/10/01 16:08:39 $
 *
 *
 * HISTORY
 *
 * $Log: WithError.h,v $
 * Revision 1.3  2002/10/01 16:08:39  s02davro
 * Bugfix of / and *
 *
 * Revision 1.3 2002/09/30 20:00:00 davpo603
 * Corrected bugs in * and /
 *
 * $Log: WithError.h,v $
 * Revision 1.3  2002/10/01 16:08:39  s02davro
 * Bugfix of / and *
 *
 * Revision 1.2  2002/08/31 14:56:49  s02davro
 * Corrected some small bugs
 *
 * Revision 1.1  2002/08/02 21:37:28  s02davro
 * Created template class to handle values with error.
 *
 *
 */

/**
 * @file WithError.h
 * Include file for the class CWithError.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: s02davro $
 * $Date: 2002/10/01 16:08:39 $
 */


#ifndef WITHERROR_H
#define WITHERROR_H

#include <string>
#include <iostream>
#include "string_utils.h"
#include "machine_dependent_types.h"
#include "math_utils.h"

/** String to use as the header when outputing a WithError value. */
#define ERROR_HEAD std::string("#Err")

RS_BEGIN_NAMESPACE

/**
 * @class WithError WithError.h
 * Template class for representing basic types that can have an error.
 * The template argument only defines what type should be annotated
 * with a possibility to have an error.
 * The only demands on the template argument T is that the arithmetic
 * functions are defined on it.
 * The dequantize (invQuatize) methods are partly stolen from 
 * UvA Trilearn (C) Remco de Boer & Jelle Kok.
 *
 * @todo Testing
 * @todo Represent the error with one value, not two.
 * @todo Check the range of errors when doing operations.
 * @todo Should the comparison operators be implemented in a different way?
 * @todo + - values in invQuantize need to be checked for sanity.
 * @todo Sin, cos and other nonlinear functions need to be handled.
 * @todo Some arithmetic functions are missing.
 *
 * @par Unit:
 *      Library
 * $Revision: 1.3 $
 * $Author: s02davro $
 * $Date: 2002/10/01 16:08:39 $
 */

enum ErrorType {
  ERROR_QUANTIZE,
  ERROR_DISTANCE,
  ERROR_DISTCHANGE,
  ERROR_NORMAL,
  ERROR_NONE
};

template <typename T> class WithError {
public:    
  /** Default constructor, no error, value set to T(). */
  WithError();

  /** Construct a object using specified value, no error. */
  WithError(const T& def_value);

  /** Construct object using a errortype according to server noise.
   *  This can be used to easy get the error interval a value from
   *  the soccerserver can be in. The third parameter is used as the
   *  quantize step, the forth as a second basevalue when calculating
   *  the distance change error. The fifth is the second quantize step
   *  in this last case. */
  WithError(const ErrorType c, const T& value, 
	    const Float64 quant_step1 = 0.1, const T& distance = T(),
	    const Float64 quant_step2 = 0.02);

  /** Construct an object using specifed value and error values. */
  WithError(const T& value, const T& plus, const T& minus);   

  /** Copyconstructor. */
  WithError(const WithError<T>& obj);

  /** @name Assignment operators. */
  //@{
  WithError<T>& operator= (const T& op);
  WithError<T>& operator= (const WithError<T>& op);
  //@}

  /** @name Arithmetic operators. */
  //@{
  WithError<T>& operator++ ();
  WithError<T>& operator-- ();
  WithError<T>& operator+= (const WithError<T>& op);
  WithError<T>& operator+= (const T& c); 
  WithError<T>& operator-= (const WithError<T>& op);
  WithError<T>& operator-= (const T& c); 
  WithError<T>& operator*= (const WithError<T>& op);
  WithError<T>& operator*= (const T& c); 
  WithError<T>& operator/= (const WithError<T>& op);
  WithError<T>& operator/= (const T& c);  
  WithError<T> operator- () const;
  WithError<T> operator+ (const WithError<T>& op) const;
  WithError<T> operator+ (const T& c) const; 
  WithError<T> operator- (const WithError<T>& op) const;
  WithError<T> operator- (const T& c) const; 
  WithError<T> operator* (const WithError<T>& op) const;
  WithError<T> operator* (const T& c) const;
  WithError<T> operator/ (const WithError<T>& op) const;
  WithError<T> operator/ (const T& c) const;
  //@}

  /** @name Comparison operators. 
   *  These functions do not care about the error values. 
   *  Maybe they should in the future. */
  //@{
  bool operator!  () const;
  bool operator== (const T& op) const;
  bool operator== (const WithError<T>& op) const;
  bool operator!= (const T& op) const;
  bool operator!= (const WithError<T>& op) const;
  bool operator<  (const T& op) const;
  bool operator<  (const WithError<T>& op) const;
  bool operator<= (const T& op) const;
  bool operator<= (const WithError<T>& op) const;
  bool operator>  (const T& op) const;
  bool operator>  (const WithError<T>& op) const;
  bool operator>= (const T& op) const;
  bool operator>= (const WithError<T>& op) const;
  //@}

  /** If the intervals represented in the two objects overlap the
   *  resulting object will be set to the intersection between this two
   *  intervals and the value as the middle.
   *  Example: 45 (2, -1), 45 (1, -0) --(improve)--> 45.5 (0.5, -0.5) */
  WithError<T>& improve (const WithError<T>& op);

  /** The resulting object will be set to the union between the two intervals
   *  Example: 45 (2, -2), 43 (2, -2) --(disimprove)--> 44 (3, -3) */      
  WithError<T>& disimprove (const WithError<T>& op);

  /** Get the value stripped from error information. */
  T getValue() const;

  /** Get the minimum value according to the error interval. */
  T getMin() const;

  /** Get the maximum value according to the error interval. */
  T getMax() const;

  /** Get the plus value.
   *  getPlus = getMax - getValue */
  T getPlus() const;

  /** Get the minus value.
   *  getMinus = getMin - getValue */
  T getMinus() const;
  
protected:
  /** Dequantize the value to get a error interval. */
  void quantizeError(const T def_value, const Float64 quant_step);

  /** Dequantize the value as a distance to get a error interval. */
  void distanceError(const T def_value, const Float64 quant_step);

  /** Dequantize as a distchange value to get a error interval. */
  void distchangeError(const T def_value, 
		       const Float64 quant_step1,
		       const T def_distance,
		       const Float64 quant_step2);

  /** Set the errors to none. */
  void noError();

  /** Normalerror for future use. Not yet implemented. */
  void normalError(const T def_value);

  /** @name Helper functions */
  //@{
  T min(T t1, T t2, T t3, T t4) const;
  T max(T t1, T t2, T t3, T t4) const;
  T min(T t1, T t2) const;
  T max(T t1, T t2) const;
  //@}

  /** @name dequantize help functions */
  //@{
  T distanceMin(const T def_value, const Float64 qstep) const;
  T distanceMax(const T def_value, const Float64 qstep) const;
  T invQMin(const T def_value, const Float64 qstep) const;
  T invQMax(const T def_value, const Float64 qstep) const;
  //@}

  /** Representation of the value. */
  T value;
  
  /** Representation of the upper bound in the error interval. */
  T plus;

  /** Representation of the lower bound in the error interval. */
  T minus;
};

template <typename T>
WithError<T>::WithError(const ErrorType c, const T& def_value, 
			const Float64 quant_step1, const T& def_distance,
			const Float64 quant_step2) 
  : value(def_value) {
  switch (c) {
  case ERROR_QUANTIZE: {
    quantizeError(def_value, quant_step1);
    break; 
  }
  case ERROR_DISTANCE: {
    distanceError(def_value, quant_step1);
    break;
  }
  case ERROR_DISTCHANGE: {
    distchangeError(def_value, quant_step1, def_distance, quant_step2);
    break;
  }
  case ERROR_NORMAL: {
    normalError(def_value);
    break;
  }
  case ERROR_NONE: {
    noError();
    break;
  }
  default: {
    noError();
    break;
  }
  }
}

template <typename T>
WithError<T>::WithError(const T& def_value) 
  : value(def_value),
    plus(),
    minus() {
}

template <typename T>
WithError<T>::WithError() 
  :value(),
   plus(),
   minus() { 
}

template <typename T>
WithError<T>::WithError(const T& def_value, const T& def_plus, 
			const T& def_minus) 
  : value(def_value),
    plus(def_plus),
    minus(def_minus) {
}

template <typename T>
WithError<T>::WithError(const WithError<T>& obj) 
  : value(obj.value),
    plus(obj.plus),
    minus(obj.minus) {
}
  
template <typename T>
WithError<T>& WithError<T>::operator= (const WithError<T>& c) {
  if ( &c != this) {
    value = c.value;
    plus = c.plus;
    minus = c.minus;
  }
  return *this;
}
  
template <typename T>
WithError<T>& WithError<T>::operator= (const T& c) {
  value = c;
  plus = T();
  minus = T();
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator++ () {
  value++;
  return *this;
}


template <typename T>
WithError<T>& WithError<T>::operator-- () {
  value++;
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator+= (const T& c) {
  value += c;
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator+= (const WithError<T>& c) {
  value += c.value;
  plus += c.plus;
  minus += c.minus;
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator-= (const T& c) {
  value -= c;
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator-= (const WithError<T>& c) {
  value -= c.value;
  plus += c.plus;
  minus += c.minus;
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator*= (const T& c) {
  value *= c;
  plus  *= c;
  minus *= c;
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator*= (const WithError<T>& c) {
  T newValue = value * c.value;
  plus = max ((value + plus) * (c.value + c.plus),
	      (value + plus) * (c.value + c.minus),
	      (value + minus) * (c.value + c.plus),
	      (value + minus) * (c.value + c.minus)) - newValue;
  minus = min ((value + plus) * (c.value + c.plus),
	       (value + plus) * (c.value + c.minus),
	       (value + minus) * (c.value + c.plus),
	       (value + minus) * (c.value + c.minus)) - newValue;
  value = newValue;
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator/= (const T& c) {
  value /= c;
  plus  /= c;
  minus /= c;
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::operator/= (const WithError<T>& c) {
  T newValue = value / c.value;
  plus = max ((value + plus) / (c.value + c.plus),
	      (value + plus) / (c.value + c.minus),
	      (value + minus) / (c.value + c.plus),
	      (value + minus) / (c.value + c.minus)) - newValue;
  minus = min ((value + plus) / (c.value + c.plus),
	       (value + plus) / (c.value + c.minus),
	       (value + minus) / (c.value + c.plus),
	       (value + minus) / (c.value + c.minus)) - newValue;
  value = newValue;
  return *this;
}


template <typename T>
WithError<T> WithError<T>::operator- () const {
  return WithError(-value, plus, minus);
}

template <typename T>
WithError<T> WithError<T>::operator+ (const T& c) const {
  return WithError<T>(value + c, plus, minus);
}

template <typename T>
WithError<T> WithError<T>::operator+ (const WithError<T>& c) const {
  return WithError<T>(value + c.value,
		      plus + c.plus,
		      minus + c.minus);
}

template <typename T>
WithError<T> WithError<T>::operator- (const T& c) const {
  return WithError<T>(value - c, plus, minus);
}

template <typename T>
WithError<T> WithError<T>::operator- (const WithError<T>& c) const {
  return WithError<T>(value - c.value,
		      plus + c.plus,
		      minus + c.minus);
}

template <typename T>
WithError<T> WithError<T>::operator* (const T& c) const {
  return WithError<T>(value * c, plus * c, minus * c);
}

template <typename T>
WithError<T> WithError<T>::operator* (const WithError<T>& c) const {
  return WithError<T>(value * c.value,
		      max ((value + plus) * (c.value + c.plus),
			   (value + plus) * (c.value + c.minus),
			   (value + minus) * (c.value + c.plus),
			   (value + minus) * (c.value + c.minus)) - 
		      (value * c.value),
		      min ((value + plus) * (c.value + c.plus),
			   (value + plus) * (c.value + c.minus),
			   (value + minus) * (c.value + c.plus),
			   (value + minus) * (c.value + c.minus)) - 
		      (value * c.value));
}

template <typename T>
WithError<T> WithError<T>::operator/ (const T& c) const {
  return WithError<T>(value / c, plus / c, minus / c);
}

template <typename T>
WithError<T> WithError<T>::operator/ (const WithError<T>& c) const {
  return WithError<T>(value / c.value,
		      max ((value + plus) / (c.value + c.plus),
			   (value + plus) / (c.value + c.minus),
			   (value + minus) / (c.value + c.plus),
			   (value + minus) / (c.value + c.minus)) - 
		      (value / c.value),
		      min ((value + plus) / (c.value + c.plus),
			   (value + plus) / (c.value + c.minus),
			   (value + minus) / (c.value + c.plus),
			   (value + minus) / (c.value + c.minus)) - 
		      (value / c.value));
}

template <typename T>
bool WithError<T>::operator! () const {
  return !value;
}

template <typename T>
bool WithError<T>::operator== (const T& op) const {
  return ((value == op.value) && 
	  (plus == T()) && 
	  (minus == T()));
}

template <typename T>
bool WithError<T>::operator== (const WithError<T>& op) const {
  return ((value == op.value) && 
	  (plus == op.plus) && 
	  (minus == op.minus));
}

template <typename T>
bool WithError<T>::operator!= (const T& op) const {
  return !(*value == op);
}

template <typename T>
bool WithError<T>::operator!= (const WithError<T>& op) const {
  return !(*this == op);
}

template <typename T>
bool WithError<T>::operator< (const T& op) const {
  return value < op;
}

template <typename T>
bool WithError<T>::operator< (const WithError<T>& op) const {
  return value < op.value;
}

template <typename T>
bool WithError<T>::operator<= (const T& op) const {
  return value <= op;
}

template <typename T>
bool WithError<T>::operator<= (const WithError<T>& op) const {
  return value <= op.value;
}

template <typename T>
bool WithError<T>::operator> (const T& op) const {
  return value > op;
}

template <typename T>
bool WithError<T>::operator> (const WithError<T>& op) const {
  return value > op.value;
}

template <typename T>
bool WithError<T>::operator>= (const T& op) const {
  return value >= op;
}

template <typename T>
bool WithError<T>::operator>= (const WithError<T>& op) const {
  return value >= op.value;
}

template<typename T>
WithError<T> operator- (const T& op1, const WithError<T>& op2) {
  return (-op2) + op1;
}

template<typename T>
WithError<T> operator+ (const T& op1, const WithError<T>& op2) {
  return op2 + op1;
}

template<typename T>
WithError<T> operator* (const T& op1, const WithError<T>& op2) {
  return op2 * op1;
}

template<typename T>
WithError<T> operator/ (const T& op1, const WithError<T>& op2) {
  return WithError<T>(op1) / op2;
}

template<typename T>
bool operator==(const T& op1, const WithError<T>& op2) {
  return op2 == op1;
}

template<typename T>
bool operator!=(const T& op1, const WithError<T>& op2) {
  return op2 != op1;
}

template<typename T>
bool operator<(const T& op1, const WithError<T>& op2) {
  return op2 >= op1;
}

template<typename T>
bool operator<=(const T& op1, const WithError<T>& op2) {
  return op2 > op1;
}

template<typename T>
bool operator>(const T& op1, const WithError<T>& op2) {
  return op2 <= op1;
}

template<typename T>
bool operator>=(const T& op1, const WithError<T>& op2) {
  return op2 < op1;
}

template <typename T>
WithError<T>& WithError<T>::improve (const WithError<T>& op) {
  if (((getMax() >= op.getMin()) && (getMin() <= op.getMax())) ||
      ((getMin() <= op.getMax()) && (getMax() >= op.getMin()))) {
    T newMax = min(getMax(), op.getMax());
    T newMin = max(getMin(), op.getMin());
    value = (newMax + newMin) / 2.0;
    plus = newMax - value;
    minus = newMin - value;
  }
  return *this;
}

template <typename T>
WithError<T>& WithError<T>::disimprove (const WithError<T>& op) {
  plus = max(getMax(), op.getMax()) - value;
  minus = min(getMin(), op.getMin()) - value;  
  value = (plus + minus)/2;
  return *this;
}
  
template <typename T>
T WithError<T>::getValue() const {
  return value;
}

template <typename T>
T WithError<T>::getMax() const {
  return value + plus;
}

template <typename T>
T WithError<T>::getMin() const {
  return value + minus;
}

template <typename T>
T WithError<T>::getPlus() const {
  return plus;
}

template <typename T>
T WithError<T>::getMinus() const {
  return minus;
}

template<typename T> 
T WithError<T>::invQMin(const T def_value, const Float64 qstep) const {
  return (rint(def_value/qstep)-0.5)*qstep;
}

template<typename T>
T WithError<T>::invQMax(const T def_value, const Float64 qstep) const {
  return (rint(def_value/qstep)+0.5)*qstep;
}

template<typename T>
T WithError<T>::distanceMin(const T def_value, const Float64 qstep) const {
  return exp ( invQMin( log (invQMin (def_value - 1.0e-10, 0.1)), qstep));
}

template<typename T>
T WithError<T>::distanceMax(const T def_value, const Float64 qstep) const {
  return exp ( invQMax( log (invQMax (def_value + 1.0e-10, 0.1)), qstep));
}

template<typename T>
void WithError<T>::quantizeError(const T def_value, const Float64 quant_step) {
  minus = invQMin(def_value, quant_step) - def_value;
  plus = invQMax(def_value, quant_step) - def_value;
}

template<typename T> 
void WithError<T>::distanceError(const T def_value, const Float64 quant_step) {
  minus = distanceMin(def_value, quant_step) - def_value;
  plus = distanceMax(def_value, quant_step) - def_value;
}

template<typename T>
void WithError<T>::distchangeError(const T def_value,
				   const Float64 quant_step1,
				   const T distance,
				   const Float64 quant_step2) {


  T dmin = distanceMin(distance, quant_step1);
  T dmax = distanceMax(distance, quant_step1);
  
  T dout = def_value/distance;
  T dmaxch = invQMin (dout, quant_step2);
  T dminch = invQMax (dout, quant_step2);
  
  minus = min(dmin*dminch, dmax*dminch) - def_value;
  plus = max(dmin*dmaxch, dmax*dmaxch) - def_value;
}

template<typename T>
void WithError<T>::normalError(const T) {
  /** Fixme! */  
  minus = T();
  plus = T();
}

template<typename T>
void WithError<T>::noError() {
  minus = T();
  plus = T();
}

template<typename T>
T WithError<T>::min(T t1, T t2, T t3, T t4) const {
  T tmp = t1;
  if (tmp > t2) tmp = t2;
  if (tmp > t3) tmp = t3;  
  if (tmp > t3) tmp = t4;
  return tmp;
}

template<typename T>
T WithError<T>::max(T t1, T t2, T t3, T t4) const {
  T tmp = t1;
  if (tmp < t2) tmp = t2;
  if (tmp < t3) tmp = t3;  
  if (tmp < t3) tmp = t4;
  return tmp;
}  

template<typename T>
T WithError<T>::min(T t1, T t2) const {
  if (t1 < t2) return t1;
  return t2;
}  

template<typename T>
T WithError<T>::max(T t1, T t2) const {
  if (t1 < t2) return t2;
  return t1;
}  

template<typename T>
std::ostream& operator<<(std::ostream& os, const WithError<T>& obj) {
#ifdef ERROR_HEAD
  if (obj.getPlus() != T() || obj.getMinus() != T())
    return os << ERROR_HEAD << "(" << toString(obj.getValue()) << " (" 
	      << toString(obj.getPlus()) << ", " 
	      << toString(obj.getMinus()) << ")";
  return os << ERROR_HEAD << "(" << toString(obj.getValue()) << ")";
#else
  return os << "(" << toString(obj.getValue()) << ")";
#endif
}

template<typename T>
const std::string toString(const WithError<T>& obj)
{
#ifdef ERROR_HEAD
  if (obj.getPlus() != T() || obj.getMinus() != T())
    return ERROR_HEAD + "(" + toString(obj.getValue()) + " (" +
      toString(obj.getPlus()) + ", " + toString(obj.getMinus()) + ")";
  return ERROR_HEAD + "(" + toString(obj.getValue()) + ")";
#else
  return "(" + toString(obj.getValue()) + ")";
#endif
}

template<typename T>
WithError<T> sqrt(const WithError<T>& op) {
  T val = sqrt(op.getValue());
  T max = val;
  T min = val;
  if (op.getMax() >= 0) 
    max = sqrt(op.getMax());
  if (op.getMin() >= 0)
    min = sqrt(op.getMin());  
  return WithError<T>(val, max - val, min - val);
}

template<typename T>
WithError<T> sqr(const WithError<T>& op) {
  T tmp = sqr(op.getValue());    
  return WithError<T>(tmp, sqr(op.getMax()) - tmp,
		      sqr(op.getMin() - tmp));
}


RS_END_NAMESPACE

#endif // !WITHERROR_H

